/******************************************************************************
*******************************************************************************
****	Copyright (c) 1997, 1998, 1999, 2000, 2001
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	????
**
**FILE NAME:
**	CnxConfig.c
**
**ABSTRACT:
**	This file implements the configuration utility for Tigris ADSL modem
**	driver.  The application will loop through the command line parameters
**	and issue IOCTL commands to the driver.
**
**DETAILS:
**	????
**
*******************************************************************************
** KEYWORDS:
**	$Archive: /Projects/Tigris and Derivatives Linux/Reference Designs/Yukon-AV/Linux_2.4_ATM_Driver/DownLoadApp/cnxadslconfig.c $
**	$Revision: 12 $
**	$Date: 5/22/01 10:12a $
*******************************************************************************
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <stdint.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <atm.h>

#include "../LnxDefs.h"
#include "../CommonData.h"
#include "../ConfigLabels.h"

// local functions
static void Usage(void);
static void Usage_Engineering(void);
static int AdjustIntParameter(char *pParameterLabelStr, char *pNewValueStr, char *pFormat);
static int ValidateInteger(char *pParameterLabelStr, char *pNewValueStr, char Foramt);
static void ClearStatistics(void);
static int ReadParameter (char *pParameterLabelStr, char *Found_Ptr);
static void ReadAllParameters (void);
static void ReadCommonParameters (void);
static void PrintHeadendCodes (void);
static int ParseParm (char *optarg, char *ParmName, char *Format, char *Value);

static struct atmif_sioc Socketioc;

#define MAX_PARM_NAME_LEN	80
#define MAX_FORMAT_LEN		80
#define MAX_VALUE_LEN		80

/******************************************************************************
*	FUNCTION NAME:
*		main
*
*	ABSTRACT:
*		This is main entrypoint for the download executable.  It is
*		responsible for parsing the command line and sending the
*		commands.
*
*	PARAMETERS:
*	    argc - command line parameter count
*	    argv - command line parameters
*
*	RETURN:
*		The status of the operation
*
*	DETAILS:
*		See http://www.skip.adb.gu.se/~nacka/glibc/glibc_479.html#SEC486
*		for details on using getopt_long
******************************************************************************/
int main(int argc,char **argv)
{
  int c;
  int value;
  NTSTATUS Status=STATUS_SUCCESS;

  // show usage if nothing is requested
  if (argc == 1)
    Usage();

  while (1)
  {
    enum {
      PARM_INDEX,
      VPI_INDEX,
      VCI_INDEX
    };

    static struct option long_options[] =
    {
      {"parm", required_argument, 0, PARM_INDEX},
      {"vpi",  required_argument, 0, VPI_INDEX},
      {"vci",  required_argument, 0, VCI_INDEX},
      {0, 0, 0, 0}
    };

    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "a:ce:f:h:i:m:p:q:w:s:u:?",
		     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c)
    {
    case PARM_INDEX:
    {
      char ParmName [MAX_PARM_NAME_LEN+1];
      char Format [MAX_FORMAT_LEN+1];
      char Value [MAX_VALUE_LEN+1];

      ParseParm (optarg, ParmName, Format, Value);

      Status = AdjustIntParameter(ParmName, Value, Format);
      break;
    }

    case VPI_INDEX:
      // its all ok so set the value in the file
      Status = AdjustIntParameter(ATM_VPI, optarg, "");
      break;

    case VCI_INDEX:
      // its all ok so set the value in the file
      Status = AdjustIntParameter(ATM_VCI, optarg, "");
      break;

    case 'q' :
      // its all ok 
      switch (*optarg)
      {
        case '0' :
          ReadCommonParameters ();
          break;

        case '1' :
          ReadAllParameters ();
          break;

        case 'h' :
          PrintHeadendCodes ();
          break;

        case 'e' :
          Usage ();
          Usage_Engineering ();
          break;

        default:
          printf("Parameter error: Query must be between 0, 1 or \"h\"\n");
	  break;
      }
      break;

    // h argument is for ADSL headend (Vendor)
    case 'h' :
      // validate that it's an integer string
      Status = ValidateInteger(ADSL_HEAD_END, optarg, 'x');
      if (Status == STATUS_SUCCESS)
      {
	// get the value and range check it

	sscanf(optarg,"%x",&value);
	if (((value < 2) || (value > 0x51)) &&
	     value &&
	     (value != 0xB6DB))
	  printf("Parameter error: %s must be 0, 0xB6DB or between 2 and 0x51\n", ADSL_HEAD_END);
	else // its all ok so set the value in the file
	  Status = AdjustIntParameter(ADSL_HEAD_END, optarg, "%lx");
      }
      break;

    case 'i' :
      Socketioc.number = atoi(optarg);
      break;

    // e argument is for ADSL headend Environment
    case 'e':
      // validate that it's an integer string
      Status = ValidateInteger(ADSL_HEAD_END_ENVIRONMENT, optarg, 'd');
      if (Status == STATUS_SUCCESS)
      {
	// get the value and range check it

	sscanf(optarg,"%d",&value);
	if (value > 1)
	  printf("Parameter error: %s must be between 0 and 1\n", ADSL_HEAD_END_ENVIRONMENT);
	else // its all ok so set the value in the file
	  Status = AdjustIntParameter(ADSL_HEAD_END_ENVIRONMENT, optarg, "");
      }
      break;

    // w argument is for wiring selection
    case 'w':
      // validate that it's an integer string
      Status = ValidateInteger(AUTO_WIRING_SELECTION, optarg, 'd');
      if (Status != STATUS_SUCCESS)
        break;

      // get the value and range check it

      sscanf(optarg,"%d",&value);
      if (value > 4)
        printf("Parameter error: %s must be between 0 and 4\n", AUTO_WIRING_SELECTION);
      else // its all ok so set the value in the file
	Status = AdjustIntParameter(AUTO_WIRING_SELECTION, optarg, "");
      break;

    // m argument is for modulation
    case 'm':
      // validate that it's an integer string
      Status = ValidateInteger(AUTO_SENSE_HANDSHAKE, optarg, 'd');
      if (Status == STATUS_SUCCESS)
      {
	// get the value and range check it
	sscanf(optarg,"%d",&value);
	if (value > 3)
	  printf("Parameter error: %s must be between 0 and 3\n", AUTO_SENSE_HANDSHAKE);
	else // its all ok so set the value in the file
	  Status = AdjustIntParameter(AUTO_SENSE_HANDSHAKE, optarg, "");
      }
      break;

    // a argument enables/disables autologging
    case 'a':
      if (!strcmp(optarg,"enable"))
        Status = AdjustIntParameter(AUTOLOG_ENABLED, "1", "");
      else if (!strcmp(optarg,"disable"))
        Status = AdjustIntParameter(AUTOLOG_ENABLED, "0", "");
      else
      {
        Usage();
        Status = EFAIL;
      }
      break;

    // f argument sets the autolog file name
    case 'f':
    {
      FILE *fptr;

      // first check to see if I can open the file for writing
      fptr = fopen(optarg, "wb");

      // if so then set the parameter
      if (fptr > 0)
        Status = AdjustIntParameter(AUTOLOG_PATH, optarg, "");
      else // otherwise error out
      {
        printf("Failure: cannot create new log file.\n");
        Status = EFAIL;
      }
      break;
    }

    // u selects the autolog update frequency
    case 'u':
      // validate that it's an integer string
      Status = ValidateInteger(AUTOLOG_UPDATE_FREQUENCY, optarg, 'd');
      if (Status == STATUS_SUCCESS) // its ok so set the value in the file
	Status = AdjustIntParameter(AUTOLOG_UPDATE_FREQUENCY, optarg, "");
      break;

    // p selects the peak cell rate in cells per second
    case 'p':
      // validate that it's an integer string
      Status = ValidateInteger(PEAK_CELL_RATE, optarg, 'd');
      if (Status == STATUS_SUCCESS) // its ok so set the value in the file
	Status = AdjustIntParameter(PEAK_CELL_RATE, optarg, "");
      break;

    // s set the max autolog file size
    case 's':
      // validate that it's an integer string
      Status = ValidateInteger(AUTOLOG_MAX_FILE_SIZE, optarg, 'd');
      if (Status == STATUS_SUCCESS) // its ok so set the value in the file
	Status = AdjustIntParameter(AUTOLOG_MAX_FILE_SIZE, optarg, "");
      break;

    // s set the max autolog file size
    case 'c':
      ClearStatistics();
      break;

    case '?':
    default:
      /* getopt_long already printed an error message. */
      Usage();
      return(1);
      break;
    }
  }

  /* Print any remaining command line arguments (not options). */
  if (optind < argc)
  {
    printf ("non-option ARGV-elements: ");
    while (optind < argc)
      printf ("%s ", argv[optind++]);
    putchar ('\n');
  }

  exit (0);
}

/******************************************************************************
*  FUNCTION NAME:
*    Usage
*
*  ABSTRACT:
*    This function prints out the usage script
*
*  PARAMETERS:
*    None
*
*  RETURN:
*    None
*
*  DETAILS:
******************************************************************************/
static void Usage()
{
  printf("Usage: cnxadslconfig --vpi=0 --vci=35 -q0 -w0 -m0 -p100000 "
   "-c -h0 -e0 -aenable -u10 -flog.txt -s100000\n"
"Used to set Conexant AccessRunner(TM) ADSL modem configuration parameters.\n"
"Note that values will not take effect until the driver is restarted.\n"
"(e.g. cnxadslctl.sh stop / cnxadslctl.sh start)\n"
"Parameter definitions:\n"
"     -i itf\n"
"  Query Parameters:\n"
"     -q Query\n"
"          0 - query current common settings\n"
"          1 - query current settings (all)\n"
"          h - query ADSL headend choices from ANSI T1.413 Annex D\n"
"\n"
"  ATM Parameters\n"
"     --vpi VPI\n"
"     --vci VCI\n"
"\n"
"  AutoSense Parameters:\n"
"     -w Wiring Selection:\n"
"          0 - automatic wiring detection\n"
"          1 - selects combo1 wiring\n"
"          2 - selects combo2 wiring\n"
"          3 - selects combo3 wiring\n"
"          4 - selects combo4 wiring\n"
"     -m Modulation:\n"
"          0 - selects Auto modulation (G.HS preferred)\n"
"          1 - selects Auto modulation (ANSI T1.413 preferred)\n"
"          2 - selects G.HS modulation\n"
"          3 - selects ANSI T1.413 modulation\n"
"\n"
"  Data Traffic Parameters:\n"
"     -p Peak Cell Rate in cells per second (0 for max)\n"
"     -c Clears transfer statistics\n"
"\n"
"  ADSL Headend Parameters:\n"
"     -h ADSL headend (vendor) code from ANSI T1.413 annex D\n"
"          0 - Non-specific\n"
"          2-51, B6DB - hex value selects specific vendor\n"
"                       (to see vendor codes, type cnxadslconfig -qh)\n"
"     -e ADSL headend environment\n"
"          0 - selects non-specific\n"
"          1 - selects no line-driver\n"
"\n"
"  Autolog Parameters:\n"
"     -a enables or disables autologing:\n"
"          enable  - enable logging\n"
"          disable - disable logging\n"
"     -f sets autolog filename\n"
"     -u sets autolog update frequency in logs per 10 seconds\n"
"     -s sets autolog max file size\n");
}

/******************************************************************************
*  FUNCTION NAME:
*    Usage
*
*  ABSTRACT:
*    This function prints out the usage script for the engineering functions
*
*  PARAMETERS:
*    None
*
*  RETURN:
*    None
*
*  DETAILS:
******************************************************************************/
static void Usage_Engineering()
{
printf(
"\n"
"  Miscellenous Parameters:\n"
"     --parm PARM=[(FORMAT)]VAL\n"
"          where PARM is any parameter name\n"
"          where FORMAT is an optional format of the form:\n"
"             %%[m]c\n"
"                 where \"m\" is the optional length modifier\n"
"                       b = byte\n"
"                       h = short int\n"
"                       l = long int (default)\n"
"                 and \"c\" is the conversion character\n"
"                       d = decimal\n"
"                       x = hex\n"
"                 default is %%ld\n"
"          where VAL is any value\n"
"          e.g. --parm \"DebugFlag=(%%lx)FFFFFFFF\"\n"
"      Note that quotes are required when using parenthesis.\n");
}

/******************************************************************************
*  FUNCTION NAME:
*    PrintHeadendCodes
*
*  ABSTRACT:
*    This function prints out the headend codes.
*
*  RETURN:
*    None
*
*  DETAILS:
******************************************************************************/
static void PrintHeadendCodes (void)
{
  printf ("ADSL headend choices from ANSI T1.413 Annex D (in hex):\n"
"0000=Non-specific\n"
"0002=Westell Inc.\n"
"0003=ECI Telecom\n"
"0004=Texas Instruments\n"
"0005=Intel\n"
"0006=Amati Communications Corp.\n"
"0007=General Data Communication\n"
"0008=Level One Communications\n"
"0009=Crystal Semiconductor\n"
"000A=AT&T - Network Systems\n"
"000B=Aware, Inc.\n"
"000C=Brooktree\n"
"000D=NEC\n"
"000E=Samsung\n"
"000F=Northern Telecom, Inc.\n"
"0010=PairGain Technologies\n"
"0011=AT&T - Paradyne\n"
"0012=Adtran\n"
"0013=INC\n"
"0014=ADC Telecommunications\n"
"0015=Motorola\n"
"0016=IBM Corp.\n"
"0017=Newbridge Networks Corp.\n"
"0018=DSC\n"
"0019=TelTrend\n"
"001A=Exar Corp.\n"
"001B=Siemens Stromberg-Carlson\n"
"001C=Analog Devices\n"
"001D=Nokia\n"
"001E=Ericsson Systems\n"
"001F=Tellabs Operations, Inc.\n"
"0020=Orckit Communications, Inc\n"
"0021=AWA\n"
"0022=Alcatel Network Systems In\n"
"0023=National Semiconductor Sys\n"
"0024=Italtel\n"
"0025=SAT\n"
"0026=Fujitsu Network Transmissi\n"
"0027=MITEL\n"
"0028=Conklin Instrument Corp.\n"
"0029=Diamond Lane\n"
"002A=Cabletron Systems, Inc.\n"
"002B=Davicom Semiconductor, Inc\n"
"002C=Metalink\n"
"002D=Pulsecom\n"
"002E=US Robotics\n"
"002F=AG Communications Systems\n"
"0030=Conexant Systems Inc.\n"
"0031=Harris\n"
"0032=Hayes Microcomputer Produc\n"
"0033=Co-optic\n"
"0034=Netspeed Inc.\n"
"0035=3-Com\n"
"0036=Copper Mountain Inc.\n"
"0037=Silicon Automation Systems\n"
"0038=Ascom\n"
"0039=Globespan Inc.\n"
"003A=ST Microelectronics\n"
"003B=Coppercom\n"
"003C=Compaq Computer Corp.\n"
"003D=Integrated Technology Expr\n"
"003E=Bay Networks, Inc.\n"
"003F=Next Level Communications\n"
"0040=Multitech Systems, Inc.\n"
"0041=AMD\n"
"0042=Sumitomo Electric\n"
"0043=Phillips M&N Electronics\n"
"0044=Efficient Networks, Inc.\n"
"0045=Interspeed\n"
"0046=Cisco Systems\n"
"0047=Tollgrade Communciations,\n"
"0048=Cayman Systems\n"
"0049=FlowPoint Corp.\n"
"004A=I.C.COM\n"
"004B=Matsushita\n"
"004C=Siemens Semiconductor\n"
"004D=Digital Link\n"
"004E=Digitel\n"
"004F=Alcatel Microelectronics\n"
"0050=Centillium Corp.\n"
"0051=Applied Digital Access, In\n"
"0052=Smart Link, Ltd.\n"
"B6DB=Medialincs (Korean)\n"
);
}

/******************************************************************************
*  FUNCTION NAME:
*    GetPadding
*
*  ABSTRACT:
*    Return a string of spaces to pad the string to the desired column
*
*  PARAMETERS:
*    SizeOfString - size of the string to be padded
*    LeftCol - Column number of start of string
*    RightCol - Column number to pad string to
*
*  RETURN:
*    Ptr to string of spaces of desired length
*
*  DETAILS:
*    Calculates the number of spaces needed to pad the given string to
*    the given column. The string will always be at least one space long
*    to ensure the output fields are separated by spaces.
*    The variable length string is actually just a variable amount of the 
*    trailing portion of a fixed length string. This string is dynamically
*    initialized so that its size can be determined at compile time. To
*    statically intialize it would have required a static constant string that
*    can not be guaranteed to be of matching length.
******************************************************************************/

#define LEFT_COL 1
#define FORMAT_COL 27
#define VALUE_COL 31
#define RIGHT_COL VALUE_COL

static char Spacing[RIGHT_COL];

static char * GetPadding (int SizeOfString, int LeftCol, int RightCol)
{
  #define min(a,b) (((a)<(b))?(a):(b))

  int Offset;

  // initialize spacing array if not already done
  if (Spacing [0] == 0)
  {
    int Index;
    for (Index = 0; Index < (sizeof (Spacing) -1); Index ++)
      Spacing [Index] = ' ';
  }

  // start with offset to last space
  Offset = sizeof(Spacing) - 1;

  // backup number of spaces desired
  Offset -= RightCol - LeftCol;

  // Then forward number of chars in string to subtract these
  Offset += SizeOfString;

  // Do not exceed past the last space (guarantee one space -
  // and do not run past end of Spacing array if string is too long)
  Offset = min (Offset, sizeof(Spacing) - 2);

  return & Spacing[Offset];
}  

/******************************************************************************
*  FUNCTION NAME:
*    ParseParm
*
*  ABSTRACT:
*    This function parses a parm command of the form PARM=[(FORMAT)]VALUE
*
*  RETURN:
*    success or failure of test
*
*  DETAILS:
******************************************************************************/
static int ParseParm (char *optarg, char *ParmName, char *Format, char *Value)
{
  int Start = 0;
  int Len;
  char *pSep;

  //*
  //* Extract PARM_NAME string
  //*

  // find end of parm name via search for "="
  pSep = strchr (&optarg[Start], '=');
  if (!pSep)
  {
    // missing mandatory "+"
    printf ("Missing \"=\" - %s\n", optarg);
    return STATUS_FAILURE;
  }

  // calc length of parm name
  Len = pSep - &optarg [Start];
  if (Len > MAX_PARM_NAME_LEN)
  {
    // name is too long
    printf ("Parameter name is too long - %s\n", optarg);
    return STATUS_FAILURE;
  }

  // copy the name
  strncpy (ParmName, &optarg[Start], Len);
  ParmName[Len] = 0;

  // calc new start of next field
  Start = Start + Len + 1;

  //*
  //* Extract optional FORMAT  string
  //*

  // Do we have an optional format (denoted by parenthesis)
  if (optarg[Start] == '(')
  {

    // calc start of actual format (past parenthesis)
    Start ++;

    // find end of format via search for closing parenthesis
    pSep = strchr (optarg, ')');
    if (!pSep)
    {
      // missing mandatory closing parenthesis
      printf ("Missing \")\" - %s\n", optarg);
      return STATUS_FAILURE;
    }

    // calc len of format
    Len = pSep- &optarg [Start];
    if (Len > MAX_FORMAT_LEN)
    {
      // format is too long
      printf ("Format is too long - %s\n", optarg);
      return STATUS_FAILURE;
    }

    // copy the format
    strncpy (Format, &optarg[Start], Len);
    Format[Len] = 0;

    // calc new start of next field
    Start = Start + Len + 1;
  }
  else
    Format [0] = 0;

  //*
  //* Extract Value string
  //*

  // find end of value via search for 0 string terminator
  pSep = strchr (&optarg[Start], 0);
  if (!pSep)
  {
    // missing mandatory 0 string terminator
    printf ("Missing value - %s\n", optarg);
    return STATUS_FAILURE;
  }

  // calc len of value
  Len = pSep - &optarg [Start];
  if (Len > MAX_PARM_NAME_LEN)
  {
    // value is too long
    printf ("Value name is too long - %s\n", optarg);
    return STATUS_FAILURE;
  }

  // copy the value
  strncpy (Value, &optarg[Start], Len);
  Value[Len] = 0;

  return STATUS_SUCCESS;
}  

/******************************************************************************
*  FUNCTION NAME:
*    AdjustParameter
*
*  ABSTRACT:
*    Changes the parameters in the config file.  It will  create a new file 
*    and copy the old file into up to the parameter to be changed and then
*    replace the parameter and copy the rest of the old file.  If the parameter
*    is never found in the old file, it will append it to the end of the new 
*    file.  If all of this was successful, it delete the old file and rename
*    the new file.
*
*  PARAMETERS:
*    pParameterLabelStr - string defining the parameter
*    pNewValueStr - new value
*    pFormat - format to be used (optional)
*
*  RETURN:
*    success or failure of operation
*
*  DETAILS:
******************************************************************************/

static int AdjustIntParameter(char* pParameterLabelStr, char* pNewValueStr, char* pFormat)
{
  int Status = STATUS_SUCCESS;
  char* IStat;
  int   OStat;
  FILE *hOldFile, *hNewFile;
  char OldFileName[512],NewFileName[512];
  char LabelLine[512],*pLabel;
  
  // make the file names
  strcpy(OldFileName, CONFIGURATION_FILE_PATH);
  strcat(OldFileName, CONFIGURATION_FILENAME);
  
  strcpy(NewFileName, CONFIGURATION_FILE_PATH);
  strcat(NewFileName, SCRATCH_CONF_NAME);
  
  // open the files
  hOldFile = fopen(OldFileName, "r");
  if (hOldFile <= 0)
  {
    printf("User params file not found %s\n",OldFileName);
    return STATUS_FAILURE;
  }
  
  hNewFile = fopen(NewFileName, "wb");
  if (hNewFile <= 0)
  {
    printf("Failure: cannot create new parameter file.\n");
    return STATUS_FAILURE;
  }
  
  // loop copying parameters from the old file to the new file
  // until the requested parameter is found
  while (Status == STATUS_SUCCESS)
  {
    // jump out on eof
    IStat = fgets(LabelLine, 256, hOldFile);
    if (!IStat)
    {
      printf("Warning! User param not found: %s. Unrecognized param added.\n", pParameterLabelStr);
      break;
    }

    // skip leading spaces
    pLabel = LabelLine;
    while ((*pLabel == ' ') ||
           (*pLabel == '\t'))
      pLabel++;

    // is this the label?
    if (strncmp(pLabel, pParameterLabelStr, strlen(pParameterLabelStr)) == 0)
    {
      // make sure we matched all of the conf file's label
      if ((isspace (pLabel[strlen(pParameterLabelStr)])) ||
           (pLabel[strlen(pParameterLabelStr)] == pFormat[0])) // this is the label so fall out of the loop
        break;
    }

    // it's a comment or the wrong label, just
    // copy it over
    OStat = fputs(LabelLine, hNewFile);
    if (!OStat)
      Status = EFAIL;
  }
  
  // put the new parameter setting in the file
  if (Status == STATUS_SUCCESS)
  {
    sprintf(LabelLine,"%s%s%s%s%s\n",
      pParameterLabelStr,    // Label
      GetPadding (strlen(pParameterLabelStr), LEFT_COL, FORMAT_COL),
      pFormat,        // Format
      GetPadding (strlen(pFormat), FORMAT_COL, VALUE_COL),
      pNewValueStr      // Value
   );

    OStat = fputs(LabelLine, hNewFile);
    if (!OStat)
      Status = EFAIL;
  }
  
  // loop copying the rest of the parameters. note skip
  // the selected parameter if it is found again
  while (Status == STATUS_SUCCESS)
  {
    // jump out on eof
    IStat = fgets(LabelLine, 256, hOldFile);
    if (!IStat)
      break;

    // skip leading spaces
    pLabel = LabelLine;
    while ((*pLabel == ' ') ||
           (*pLabel == '\t'))
      pLabel++;

    // is this the label?
    if (strncmp(pLabel, pParameterLabelStr, strlen(pParameterLabelStr)) == 0)
    {
      // make sure we matched all of the conf file's label
      if ((isspace (pLabel[strlen(pParameterLabelStr)])) ||
           (pLabel[strlen(pParameterLabelStr)] == pFormat[0])) // ignore this line
        continue;
    }

    // it's a comment or the wrong label, just
    // copy it over
    OStat = fputs(LabelLine, hNewFile);
    if (!OStat)
      Status = EFAIL;
  }
  
  // close the files
  fclose(hNewFile);
  fclose(hOldFile);
  
  // if all clean to this point
  // delete the old file
  // rename the new file to the old name
  if (Status == STATUS_SUCCESS)
  {
    unlink(OldFileName);
    rename(NewFileName, OldFileName);
  }
  else  // otherwise delete the new file
    unlink(NewFileName);
  
  return Status;
}

/******************************************************************************
*  FUNCTION NAME:
*    ValidateInteger
*
*  ABSTRACT:
*    validates that the user requested value is an integer string
*
*  RETURN:
*    success or failure of test
*
*  DETAILS:
******************************************************************************/
int ValidateInteger(char *pParameterLabelStr, char *pNewValueStr, char Format)
{
  int Status = STATUS_SUCCESS;
  
  // skip leading spaces
  while(*pNewValueStr == ' ')
    pNewValueStr++;

  // loop through all chars until an EOF
  do
  {
    switch (Format)
    {
      case 'd' :
        if ((*pNewValueStr < '0') || (*pNewValueStr > '9'))
        {
          printf("Parameter failure: %s must be an integer value\n",pParameterLabelStr);
          Status = EFAIL;
          break;
        }
        break;

      case 'x' :
        if (((*pNewValueStr < '0') || (*pNewValueStr > '9')) &&
            ((*pNewValueStr < 'a') || (*pNewValueStr > 'f')) &&
            ((*pNewValueStr < 'A') || (*pNewValueStr > 'F')))
        {
    
          printf("Parameter failure: %s must be a hexadecimal value\n", pParameterLabelStr);
          Status = EFAIL;
          break;
        }
        break;
    }

    pNewValueStr++;
  } while (*pNewValueStr != 0);

  return Status;
} 

/******************************************************************************
*  FUNCTION NAME:
*    ClearStatistics
*
*  ABSTRACT:
*    This function sends an ioctl to driver to zero out statistics
*
*  PARAMETERS:
*    None
*
*  RETURN:
*    None
*
*  DETAILS:
******************************************************************************/
static void ClearStatistics(void)
{
  int SocketID;
  char Dummy[8];
  
  // open the socket 
  SocketID = socket(PF_ATMPVC, SOCK_DGRAM, 0);
  if (SocketID < 0) 
  {
    printf("Socket open failure\n");
    return;
  }

  // build and issue the command
  Socketioc.arg    = Dummy; // Linux doesn't like null here
  Socketioc.length = sizeof (Dummy);     // so I am sending fake stuff

  ioctl(SocketID, TIG_CLEAR_STATS, &Socketioc);
   
  // close the socket
  close(SocketID);
}

/******************************************************************************
*  FUNCTION NAME:
*    ReadAllParameters
*
*  ABSTRACT:
*    Read and print the All parameters.
*
*  RETURN:
*    NA
*
*  DETAILS:
******************************************************************************/
static void ReadAllParameters (void)
{
  int Status = STATUS_SUCCESS;
  char* IStat;
  FILE *hOldFile;
  char LabelLine[512];
  char *pLabel;

  // open the file
  hOldFile = fopen(CONFIGURATION_FILE_PATH CONFIGURATION_FILENAME, "r");
  if (hOldFile <= 0)
  {
    printf("User params file not found %s\n", CONFIGURATION_FILE_PATH CONFIGURATION_FILENAME);
    return;
  }

  // loop copying parameters from the old file to the new file
  // until the requested parameter is found
  while (Status == STATUS_SUCCESS)
  {
    // jump out on eof
    IStat = fgets(LabelLine, 256, hOldFile);
    if (!IStat)
      break;

    // skip leading spaces
    pLabel = LabelLine;
    while ((*pLabel == ' ') ||
           (*pLabel == '\t'))
      pLabel++;

    printf ("%s", pLabel);
  }

  // close the files
  fclose(hOldFile);
}

/******************************************************************************
*  FUNCTION NAME:
*    ReadCommonParameters
*
*  ABSTRACT:
*    Read and print the supported parameters.
*
*  RETURN:
*    NA
*
*  DETAILS:
******************************************************************************/
static void ReadCommonParameters (void)
{
  char Conf_Line [512];
  char *List [] =
  {
    AUTO_WIRING_SELECTION,
    AUTO_SENSE_HANDSHAKE,
    ADSL_HEAD_END_ENVIRONMENT,
    ADSL_HEAD_END,
    AUTOLOG_ENABLED,
    AUTOLOG_PATH,
    AUTOLOG_UPDATE_FREQUENCY,
    PEAK_CELL_RATE,
    AUTOLOG_MAX_FILE_SIZE,
    ATM_VPI,
    ATM_VCI,
    NULL
  };
  int Index;

  for (Index = 0; List[Index]; Index++)
    if (ReadParameter (List[Index], Conf_Line) == STATUS_SUCCESS)
      printf ("%s", Conf_Line);
}

/******************************************************************************
*  FUNCTION NAME:
*    ReadParameter
*
*  ABSTRACT:
*    Read and print the given parameter.
*
*  RETURN:
*    success or failure of operation
*
*  DETAILS:
******************************************************************************/
static int ReadParameter (char *pParameterLabelStr, char *Found_Ptr)
{
  int Status = STATUS_SUCCESS;
  char* IStat;
  FILE *hOldFile;
  //char OldFileName[512];
  char LabelLine[512];
  char *pLabel;

  *Found_Ptr = 0;

  // open the file
  hOldFile = fopen(CONFIGURATION_FILE_PATH CONFIGURATION_FILENAME, "r");
  if (hOldFile <= 0)
  {
    printf("User params file not found %s\n", CONFIGURATION_FILE_PATH CONFIGURATION_FILENAME);
    return STATUS_FAILURE;
  }
  
  // loop copying parameters from the old file to the new file
  // until the requested parameter is found
  while (Status == STATUS_SUCCESS)
  {
    // jump out on eof
    IStat = fgets(LabelLine, 256, hOldFile);
    if (!IStat)
    {
      printf("User param not found %s\n", pParameterLabelStr);
      break;
    }

    // skip leading spaces
    pLabel = LabelLine;
    while ((*pLabel == ' ') ||
           (*pLabel == '\t'))
      pLabel++;

    if (!strncmp(pLabel, pParameterLabelStr, strlen(pParameterLabelStr)))
    {
      // this is the label so fall out of the loop
      strcpy (Found_Ptr, pLabel);
      break;
    }
  }

  // close the files
  fclose(hOldFile);

  return Status;
}
