/******************************************************************************
*******************************************************************************
****	Copyright (c) 1997, 1998, 1999, 2000, 2001
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	????
**
**FILE NAME:
**	ARMAbstract.c
**
**ABSTRACT:
**	This file implements the ARM abstraction layer which is responsible
**	for the interface to the ARM processor.
**
**DETAILS:
**		
**
*******************************************************************************
** KEYWORDS:
**	$Archive: /Projects/Tigris and Derivatives Linux/Reference Designs/Tigris/Linux_2.2.16_ATM_Driver/KernelModule/ARMAbstract.c $
**	$Revision: 17 $
**	$Date: 7/03/01 4:39p $
*******************************************************************************
******************************************************************************/
#include "linuxmodver.h"

#include "LnxTools.h"
#include "../CommonData.h"
#include "../hexformat.h"

#ifndef _DATAPUMP_H_
	#define _DATAPUMP_H_
	typedef struct DATAPUMP * PDATAPUMP_CONTEXT;
#endif

#include "CnxADSL.h"
#include "ChipALCdsl.h"
#include "ChipALCdslV.h"
#include "ChipALInterf.h"
#include "ARMAbstract.h"

#if 1/*CLN*/
#define DBG_REPORT(msg) \
    printk(DEV_LABEL "=>%s:%i %s %s\n", __FILE__, __LINE__, __FUNCTION__, msg)
#else
#define DBG_REPORT(msg) do {} while (0)
#endif

////////////////////////////////////
// local defines
////////////////////////////////////
#define 	JUMP_ADDRESS		0x00010480
#define		EXT_COMMAND_ADDRESS	0x00010504

#define		DOWNLOAD_TIMEOUT	1000 // one second

#define IS_CMD_LOADED	(0x00000102 << NUM_ARM_RESP_BITS)

#define SSCAN		8	// GPIO8

/******************************************************************************
Private Function Prototypes
******************************************************************************/

typedef struct {
  CSR_COMMAND_PARAM_T	CsrParam;
  WORD			Size;
} GPIO_INIT_T;

typedef struct {
  CDSL_ADAPTER_T	*pAdapter;
  DWORD			Resp2;
  EVENT_HNDL		*pEvent;
} ARM_ASYC_IO_T;

//////////////////////////////////////////////////////////////////////
// Local functions
//////////////////////////////////////////////////////////////////////
LOCAL DWORD atoh(
    IN const char	*cInStr,
    IN DWORD	uNumChars
);

LOCAL NTSTATUS processIntel32Rec(
    IN CDSL_ADAPTER_T	*pThisAdapter,
    IN DWORD		InRecLength,
    IN HEX_RECORD_T	*pInRec,
    IN DWORD		uExtAddr
);

LOCAL VOID ARMALP46CompletionRoutine(
    IN PVOID		Context
);

LOCAL NTSTATUS ARMALStartArm(
    IN CDSL_ADAPTER_T	*pThisAdapter
);

NTSTATUS ARMALInitIo(
    IN CDSL_ADAPTER_T	*pThisAdapter
);

LOCAL VOID ARMALNotifyArmReady(
    IN CDSL_ADAPTER_T	*pThisAdapter
);

///////////////////////////////////////////////////////////////////////

/*++
    ARMALCfgInit

Routine Description:
    The card management module calls this function if called to
    initialize any user configurable settings.  These settings
    are passed into the driver via IOCTL commands and will be
    accessible from the adapter structure.

Arguments:
    None

Return Value:
    The status of the operation.
--*/
NDIS_STATUS ARMALCfgInit(
    I_O CDSL_ADAPTER_T	*pThisAdapter,
    IN PTIG_USER_PARAMS  pUserParams) {
  NTSTATUS Status = STATUS_SUCCESS;

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALCfgInit  %p", pThisAdapter));

  pThisAdapter->ArmALMod = NULL;

  // Allocate memory for the buffer management adapter context structure.
#if 0
  {
    ARM_ABSTRACT_T* pARMCtxt;

#define MEMORY_TAG 'ArmL'
    Status = ALLOCATE_MEMORY(pARMCtxt, sizeof(*pARMCtxt));
    if (Status == STATUS_SUCCESS) {
      CLEAR_MEMORY(pARMCtxt, sizeof(*pARMCtxt));
      pThisAdapter->ArmALMod = pARMCtxt;
    } else
      Status = -ENOMEM;
  }
#endif

  return Status;
}

/*++
    ARMALInit

Routine Description:
    The card management module calls this function to initialize 
    the ARM abstraction module.  It will initialize any data required 
    by the module and prepare the module for operations. 

Arguments:
    None

Return Value:
    The status of the operation.
--*/
NDIS_STATUS ARMALInit(
    IN CDSL_ADAPTER_T	*pThisAdapter) {
  NTSTATUS Status = STATUS_SUCCESS;

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALInit  %p", pThisAdapter));

  return Status;
}

/*++
    ARMALHalt

Routine Description:
    The card management module calls this function to terminate the 
    ARM abstraction module.  It will stop any active operations and 
    free any data allocated by the module.

Arguments:
    None

Return Value:
    The status of the operation.
--*/
NDIS_STATUS ARMALHalt(
    IN CDSL_ADAPTER_T	*pThisAdapter) {
  NTSTATUS Status = STATUS_SUCCESS;

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALHalt  %p", pThisAdapter));

  // free the buffer management adapter context structure.
#if 0 // not used
  FREE_MEMORY(pThisAdapter->ArmALMod, , 0);
#endif

  return Status;
}

static LOG_ENTRY_T Log_Buffer[NUM_LOG_ENTRIES];/*CLN*/
#if 0
static DWORD Log_Put_Index;
static BOOLEAN Log_TraceAll = FALSE;/*CLN*/

LOCAL void printl(const char *pFormat, ...) {
  va_list va_ptr;

  va_start(va_ptr, pFormat);

  if (Log_TraceAll)
    if (Log_Put_Index < NUM_LOG_ENTRIES) {
      int Index;

      Log_Buffer[Log_Put_Index].pFormat = (char *)pFormat;

      for (Index = 0; Index < NUM_LOG_PARMS; Index++)
	Log_Buffer[Log_Put_Index].Data[Index] = va_arg(va_ptr, DWORD);

      Log_Put_Index++;
    }

  va_end(va_ptr);
}
#endif

#define NUM_PRINTL_STRINGS 100

typedef struct {
  const char *pSrc;
  char *pDest;
} STRING_TABLE_T;

static STRING_TABLE_T String_Table[NUM_PRINTL_STRINGS];
static DWORD	String_Offset;
static int	String_Cnt;/*CLN*/

/******************************************************************************
FUNCTION NAME:
	Log_Copy_String

ABSTRACT:
	Copy the string if it is uniquely new.

RETURN:
	Returns a ptr to the destination buffer
	In case of overflow, it returns a ptr to first string (presumebly an
	error string)

DETAILS:
	Keep a table of ptrs to strings already copied and where they were
	copied to.
	When a new copy is requested, this table is searched and if the string
	has already been copied, a ptr to it is returned.

	If not, the string is copied to the string buffer and this is entered
	into the table if there is room. If there is no room, a ptr to the
	first entry in the table is returned.

	You should initially call this routine with an error string as this
	will be the string returned in case of buffer overflow.
******************************************************************************/
LOCAL char * Log_Copy_String(
    char *pString_Buffer,
    DWORD String_Buffer_Size,
    const char *pFormat) {
  int		String_Index;

  // search all existing copied strings for a match
  for (String_Index = 0; String_Index < String_Cnt; String_Index++)
    if (pFormat == String_Table[String_Index].pSrc)
      break;

  // did we NOT find the string?
  if (String_Index == String_Cnt) {
    if (String_Cnt <= NUM_PRINTL_STRINGS) {// There is room to add another str?
      // Add a new lookup table entry for this unique string
      String_Table[String_Cnt].pSrc = pFormat;
      String_Table[String_Cnt].pDest = &pString_Buffer[String_Offset];
      String_Cnt++;

      // copy string without exceeding size of dest buffer
      strncat(&pString_Buffer[String_Offset], pFormat, String_Buffer_Size - String_Offset);

      // advance to next string in dest buffer
      String_Offset += strlen(&pString_Buffer[String_Offset]) + 1; // +1 for \0 terminator
    } else {
      printk("Log_Put String Table overflow\n");
      String_Index = 0;
    }
  }

  // return a ptr to the copied string
  return String_Table[String_Index].pDest;
}

/******************************************************************************
FUNCTION NAME:
	Log_Scan_For_String

ABSTRACT:
	scan format for %[n]s and return which arg it is

RETURN:
	Arg index of match. -1 if string exhausted

DETAILS:
******************************************************************************/
LOCAL int Log_Scan_For_String(int Last_Arg_Found_Index, char *pFormat) {
  int Char_Index;
  int Next_Arg_Found_Index = -1;
  int Found_Match = 0; 

  typedef enum {
    LOOK_FOR_PERCENT,
    LOOK_FOR_CONV
  } STATE_T;

  STATE_T State = LOOK_FOR_PERCENT;

  for (Char_Index = 0; !Found_Match && pFormat[Char_Index]; Char_Index++) {
    // depending on state
    switch (State) {
    case LOOK_FOR_PERCENT :
      if (pFormat[Char_Index] == '%')
	State = LOOK_FOR_CONV;
      break;

    case LOOK_FOR_CONV :
      if (pFormat[Char_Index] == '%') {
	// ignore the %% pair
	// return to looking for percent
	State = LOOK_FOR_PERCENT;
      } else if ((pFormat[Char_Index] >= '0') &&
		  (pFormat[Char_Index] <= '9')) {
	// this is a length modifier - ignore
      } else if (pFormat[Char_Index] == 's') {
	// found a conversion specifier - count it
	Next_Arg_Found_Index++;

	if (Next_Arg_Found_Index > Last_Arg_Found_Index) {
	  // Found next string conversion character
	  // Done - exit loop
	  Found_Match = 1;
	} else {
	  // return to looking for percent
	  State = LOOK_FOR_PERCENT;
	}
      }
      // a specifier that wasn't %s 
      else {
	// found a conversion specifier - count it
	Next_Arg_Found_Index++;

	// return to looking for percent
	State = LOOK_FOR_PERCENT;
      }
      break;
    }
  }

  // return the parm index of the match if one was found
  if (!Found_Match)
    Next_Arg_Found_Index = -1;

  return Next_Arg_Found_Index;
}	

/******************************************************************************
FUNCTION NAME:
	Log_Copy

ABSTRACT:
	Copy the formats and data to the destination buffer

RETURN:
	NA

DETAILS:
	Copy a buffer of LOG_ENTRY types (ptr to format and N parms for printing)
	to the requested destination buffer. It also copies a buffer of format
	strings that the LOG_ENTRY will refer to.
	To save space, the buffer of format strings will re-use format strings
	rather than copy each reference to the same string.
******************************************************************************/
static DWORD Cumulative_Start_Index = 0;/*CLN*/

LOCAL void Log_Copy(TIG_PRINT_DESC *pDesc) {
  DWORD	Src_Index;
  DWORD	Dest_Index;
  DWORD	Start;

  Start = (pDesc->Cumulative) ? 0 : Cumulative_Start_Index;

  pDesc->pString_Buffer[0] = 0;
  String_Offset = 0;
  String_Cnt = 0;

  // add an error message string for printl to use when the table overflows
  Log_Copy_String(pDesc->pString_Buffer, pDesc->String_Buffer_Size, "Log_Put Format Table Overflow\n");

  // do for each log entry - limited to size of smaller of source or dest table
  for (Src_Index = Start, Dest_Index = 0; Src_Index < NUM_LOG_ENTRIES; Src_Index++, Dest_Index++) {
    if (Dest_Index >= pDesc->Buff_Size) {
      printk("Log_Copy Buffer Overflow\n");
      break;
    }

    // copy entry (includes terminating null entry)
    pDesc->pBuff[Dest_Index] = Log_Buffer[Src_Index];

    // if terminating null entry
    if (!Log_Buffer[Src_Index].pFormat)
      break;

    // add this format to string table and adjust ptr to string in table
    pDesc->pBuff[Dest_Index].pFormat = Log_Copy_String(pDesc->pString_Buffer,
						       pDesc->String_Buffer_Size,
						       Log_Buffer[Src_Index].pFormat);

    // scan format string for %[n]s that indicates ptrs to strings that must be copied
    {
      int Last_Arg_Found_Index = -1;

      do {
	Last_Arg_Found_Index = Log_Scan_For_String(Last_Arg_Found_Index, Log_Buffer[Src_Index].pFormat);
	if (Last_Arg_Found_Index >= 0) {
	  // add this string to string table and adjust ptr to string in table
	  pDesc->pBuff[Dest_Index].Data[Last_Arg_Found_Index] = (DWORD) Log_Copy_String
	    (pDesc->pString_Buffer,
	     pDesc->String_Buffer_Size,
	     (char *)Log_Buffer[Src_Index].Data[Last_Arg_Found_Index]);
	}
      } while (Last_Arg_Found_Index >= 0);
    }
  }

  // remember how far we got so we can start here next time if desired
  Cumulative_Start_Index = Src_Index;
}

/*++
    ARMALIoctl

Routine Description:
    The card management module calls this function to process commands
    received from the application.  Supported commands will include:

    * TIG_LOAD_START -   called to prepare the device for download.  No args.
    * TIG_LOAD_RECORD -  called to issue a firmware record to the device.  Arg 
                         will be a pointer to a firmware record.
    * TIG_LOAD_DONE -    called after the final record has been sent. No args.

Arguments:
    None

Return Value:
    The status of the operation.
--*/
NDIS_STATUS ARMALIoctl(
    I_O CDSL_ADAPTER_T	*pThisAdapter,
    IN unsigned int	Command,
    I_O PVOID		CmdArg) {
  NTSTATUS Status = STATUS_SUCCESS;

  // switch on the command type
  switch (Command) {
  case TIG_GET_PRINT_BUFF:
    {
      const TIG_PRINT_DESC *myCmdArg = CmdArg;
      TIG_PRINT_DESC Desc;

      if(copy_from_user(&Desc.Buff_Size,&myCmdArg->Buff_Size,
			sizeof(Desc.Buff_Size))){
	Status=-EFAULT;
	break;
      }

      Log_Copy(&Desc);

      if(copy_to_user(CmdArg, &Desc, sizeof(Desc)))
	Status=-EFAULT;
    }
    break;

  case TIG_GET_DEBUG_DATA :
    {
      TIG_DEBUG_DATA_DESC Desc;

      CLEAR_MEMORY(&Desc, sizeof(Desc));

      // load debug parms into Desc.Data[0] to [9]

      if(copy_to_user(CmdArg, &Desc, sizeof(Desc)))
	Status=-EFAULT;
    }
    break;

  case TIG_LOAD_START:
    DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		     ChipALDebugFlag,
		     (DEV_LABEL "=>ARMAbstract:ARMALIoctl  TIG_LOAD_START"));
    pThisAdapter->Loading = TRUE;

    //*
    //* See if Extended Command set is already loaded.  This will reset the
    //* AutoDownload flag if the ARM responds to any extended commands.
    //*
    ARMIsCommandSetLoaded(pThisAdapter);

    if(!pThisAdapter->AutoDownload) //* We do not need to download ARM
      Status = -EALREADYLOADED;
    break;

  case TIG_LOAD_RECORD:
    if (!pThisAdapter->Loading) {
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:ARMALIoctl  load record while not in load"));
      Status = -EBADFD;
    } else {
      const TIG_LOAD_RECORD_T *myCmdArg=CmdArg;
      TIG_LOAD_RECORD_T myRecord;

      if(copy_from_user(&myRecord,myCmdArg,sizeof(myRecord))){
	Status=-EFAULT;
	break;
      }

      switch (myRecord.RecordType) {
      case DATA_REC:
	//	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
	//		         ChipALDebugFlag,
	//		         (DEV_LABEL "=>ARMAbstract:ARMALIoctl  TIG_LOAD_RECORD RecordType=DATA_REC"));
	// call a function to process the data record
	{
	  char myData[sizeof(HEX_RECORD_T)+16*2/*MAX_RECORD_LENGTH*/];
	  HEX_RECORD_T *myRecordData=(void *)&myData[0];

	  if(myRecord.RecordLength>sizeof(myData)){
	    Status=-EFAULT;
	    break;
	  }
	  if(copy_from_user(myRecordData,myCmdArg->RecordData,
			    myRecord.RecordLength)){
	    Status=-EFAULT;
	    break;
	  }

	  Status = processIntel32Rec(pThisAdapter,
				     myRecord.RecordLength,
				     myRecordData,
				     myRecord.RecordAddress);
	}
	break;

      case EXT_GOTO_CMD_REC:
	{
	  P46_IO_COMMAND_T	ArmCommand;

	  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
			   ChipALDebugFlag,
			   (DEV_LABEL "=>ARMAbstract:ARMALIoctl  TIG_LOAD_RECORD RecordType=EXT_GOTO_CMD_REC"));

	  //*
	  //* Start Segment Address (GOTO)
	  //*
	  ArmCommand.CsrParam.NotifyOnComplete = NULL;
	  ArmCommand.CsrParam.Command = ARM_GOTO;
	  ArmCommand.CsrParam.Param.Goto.Address = myRecord.RecordAddress;

	  Status = ChipALWrite(pThisAdapter,
			       P46_ARM,
			       ADDRESS_ARM_BAR_0,
			       CSR_ARM_CMD,
			       sizeof(ArmCommand),
			       (DWORD) &ArmCommand,
			       0);
	  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
			   ChipALDebugFlag,
			   (DEV_LABEL "=>ARMAbstract:ARMALIoctl  TIG_LOAD_RECORD EXT_GOTO_CMD_REC, Status=%lx", Status));
	}
	break;

      default:
	DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			 DBG_FLAG_DONT_CARE,
			 (DEV_LABEL "=>ARMAbstract:ARMALIoctl  TIG_LOAD_RECORD - invalid load record type"));
	Status = -EINVALIDTYPE;
	break;
      }
    }
    break;

  case TIG_LOAD_DONE:
    DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		     ChipALDebugFlag,
		     (DEV_LABEL "=>ARMAbstract:ARMALIoctl  TIG_LOAD_DONE"));
    if (!pThisAdapter->Loading) {
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:ARMALIoctl  load done while not in load"));
      Status = -EBADFD;
    } else {
      Status = ARMALStartArm(pThisAdapter);
      if (Status == STATUS_SUCCESS) {
	Status = ARMALInitIo(pThisAdapter);
	if (Status == STATUS_SUCCESS)
	  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
			   ChipALDebugFlag,
			   (DEV_LABEL "=>ARMAbstract:ARMALIoctl  ARM Started and IO Programmed"));
	else
	  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			   DBG_FLAG_DONT_CARE,
			   (DEV_LABEL "=>ARMAbstract:ARMALIoctl  >ARM Not Started, ARMALInitIo failed!"));
      } else
	DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			 DBG_FLAG_DONT_CARE,
			 (DEV_LABEL "=>ARMAbstract:ARMALIoctl  >ARM Not Started, ARMALStartArm failed!"));
    }
    break;
  }

  return Status;
}

/******************************************************************************
FUNCTION NAME:
	ARMSyncRoutine

ABSTRACT:
	Used as	a general purpose completion routine so it can signal an ARM 
	response to a specific command.

RETURN:
	VOID

DETAILS:
******************************************************************************/
LOCAL VOID ARMSyncRoutine(
    I_O	void			* pSync,
    IN	RESPONSE_T		* Response) {
  ARM_ASYC_IO_T *pASync = pSync;

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMSyncRoutine"));

  // Store Response
  pASync->Resp2 = *((DWORD *)&Response->Param[0]);

  // Signal Waiting Thread
  SET_EVENT(pASync->pEvent);
}

/******************************************************************************
FUNCTION NAME:
	ARMIsCommandSetLoaded

ABSTRACT:
	Determine if ARM is already Downloaded

RETURN:
	NTSTATUS

DETAILS:
******************************************************************************/
NTSTATUS ARMIsCommandSetLoaded(
    I_O CDSL_ADAPTER_T	*pThisAdapter) {
  NTSTATUS			Status;
#if 0
  signed long 			NsecDelay = DOWNLOAD_TIMEOUT;
#endif
  int				TimeRemain;
  EVENT_HNDL			CompletionEvent;
  static volatile ARM_ASYC_IO_T	ArmSync;/*CLN*/

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMIsCommandSetLoaded"));

  // prepare the wait entry
  INIT_EVENT(&CompletionEvent);

  ArmSync.pAdapter = pThisAdapter;
  ArmSync.pEvent = &CompletionEvent;

  // add a DPC handler to get the response
  //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
  //		         ChipALDebugFlag,
  //		         (DEV_LABEL "=>ARMAbstract:ARMIsCommandSetLoaded  add event handler"));
  Status = ChipALAddEventHandler(pThisAdapter,
				 ARM_FUNCTION,
				 HW_COMMAND_SYNC,
				 HW_DPC_EVENT,
				 ARMSyncRoutine,
				 (void *)&ArmSync);
  ArmSync.Resp2 = 0;

#if 0
  // don't do this as it causes an interrupt on the DSL function before the
  // ARM function completes all operations and the Qube doesn't like
  // interrupts on both functions simultaneously.
 {
   P46_IO_COMMAND_T		ArmCommand;

   ArmCommand.CsrParam.Command = IS_CMD_LOADED | ARM_COMMAND_SYNC;
   ArmCommand.CsrParam.Param.CommandSync.Response = RESPONSE_YES;
   ArmCommand.CsrParam.NotifyOnComplete = NULL;

   // issue the command
   //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
   //		         ChipALDebugFlag,
   //		         (DEV_LABEL "=>ARMAbstract:ARMIsCommandSetLoaded  issue a write IS_CMD_LOADED"));

   //			calP46WriteCsr(pContext, BaseAddress, &ArmCommand);
   //Status = ChipALWrite(pThisAdapter,
   //                     P46_ARM,
   //                     ADDRESS_ARM_BAR_0,
   //                     CSR_ARM_CMD,
   //                     sizeof(ArmCommand),
   //                     (DWORD) &ArmCommand,
   //                      0);

   //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
   //		         ChipALDebugFlag,
   //		         (DEV_LABEL "=>ARMAbstract:ARMIsCommandSetLoaded  wait for complete"));
 }
#endif

  // speed up timeout that we know will happen since we commented out the
  // write above.
  // wait on response
  TimeRemain = WAIT_EVENT_TIMEOUT(&CompletionEvent, 1/*NsecDelay*/);

  //* Remove Sync Handler
  //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
  //		         ChipALDebugFlag,
  //		         (DEV_LABEL "=>ARMAbstract:ARMIsCommandSetLoaded  remove event handler"));
  ChipALRemoveEventHandler(pThisAdapter,
			   ARM_FUNCTION,
			   HW_COMMAND_SYNC,
			   HW_DPC_EVENT,
			   ARMSyncRoutine);

  // Make sure we did not timeout and the ARM sent a command sync
  // response.  0 indicates timeout.  It really doesn't matter if it was
  // our command sync response or if it was one from some other module.
  // We only want to know if the ARM is responding.
  if (!TimeRemain) {
    DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		     ChipALDebugFlag,
		     (DEV_LABEL "=>ARMAbstract:ARMIsCommandSetLoaded  download required"));
    pThisAdapter->AutoDownload = TRUE;
    Status = STATUS_FAILURE;
  } else {
    DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		     ChipALDebugFlag,
		     (DEV_LABEL "=>ARMAbstract:ARMIsCommandSetLoaded  no need to download"));
    pThisAdapter->AutoDownload = FALSE;
    Status = STATUS_SUCCESS;
  }

  return(Status);
}

/******************************************************************************
FUNCTION NAME:
	processIntel32Rec

ABSTRACT:
	Process Data Record

RETURN:
	NTSTATUS:

DETAILS:
	This Function was orginally based on the RipTide processDataRec 
	function. Essentially, it is the same.  The major difference is how
	writes to the hardware are performed.

	In this implementation, the ARM CSR is interrupt driven.  Waits are
	used to ensure we have one command ready to go as soon as the ARM
	finishes processing the current command.  In other words, the first
	command we send down gets written immediately.  The second command
	will be queued by the lower layer.  The second command is written
	automatically by the lower layer DPC when the ARM interrupts to
	indicate command complete.
	When written, it signals the local thread to continue processing.
******************************************************************************/
LOCAL NTSTATUS processIntel32Rec(
    IN CDSL_ADAPTER_T	*pThisAdapter,
    IN DWORD		InRecLength,
    IN HEX_RECORD_T	*pInRec,
    IN DWORD		uExtAddr) {
  DWORD 		uByteCnt;	// Number of data bytes	in this	record
  NTSTATUS		Status;		// Misc completion Status
  signed long 		NsecDelay = DOWNLOAD_TIMEOUT;
  int			TimeRemain;
  EVENT_HNDL		CompletionEvent;
  static P46_IO_COMMAND_T	ArmCommand;/*CLN*/// Command written to Hardware
  const char		* pInPtr;	// Points into the input data record

  //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
  //		         ChipALDebugFlag,
  //		         (DEV_LABEL "=>ARMAbstract:processIntel32Rec"));

  // prepare the wait entry
  INIT_EVENT(&CompletionEvent);

  // Parse the record byte count
  uByteCnt = atoh(pInRec->hdr.Length, sizeof(pInRec->hdr.Length));

  // Calculate the data load address
  uExtAddr += atoh(pInRec->hdr.Offset, sizeof(pInRec->hdr.Offset));

#if 1/*CLN*/
  if (ArmCommand.CsrParam.Command)
    DBG_REPORT("Wo");
#endif

  ArmCommand.CsrParam.NotifyOnComplete = ARMALP46CompletionRoutine;
  ArmCommand.CsrParam.UserContext = &CompletionEvent;
  ArmCommand.CsrParam.Command = ARM_SET_MEMORY_WRITE_ADDRESS;
  ArmCommand.CsrParam.Param.SetWriteAddress.Address = uExtAddr;
  ArmCommand.CsrParam.Param.SetWriteAddress.AutoIncrement = ARM_AUTO_INCREMENT_YES;

  //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
  //		         ChipALDebugFlag,
  //		         (DEV_LABEL "=>ARMAbstract:processIntel32Rec  Write Address cmd"));

  Status = ChipALWrite(pThisAdapter,
		       P46_ARM,
		       ADDRESS_ARM_BAR_0,
		       CSR_ARM_CMD,
		       sizeof(ArmCommand), // Length is unused for ARM.
		       (DWORD) &ArmCommand,
		       0);	// Mask	Value is unused	by ARM

  // wait on response
  TimeRemain = WAIT_EVENT_TIMEOUT(&CompletionEvent, NsecDelay);
  if (!TimeRemain) {
    DBG_REPORT("Write Address cmd timed out");
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>ARMAbstract:processIntel32Rec  Write Address cmd timed out"));
    Status = -ETIMEOUT;
    return (Status);
  }

  //* Set Arm Command for Dword writes
  ArmCommand.CsrParam.Command = ARM_WRITE_MEMORY;
  ArmCommand.CsrParam.Param.WriteMemory.AccessType = ARM_ACCESS_32;

  //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
  //		         ChipALDebugFlag,
  //		         (DEV_LABEL "=>ARMAbstract:processIntel32Rec  send record data"));

  // Point to the start of the data in the input record
  pInPtr = &pInRec->data[0];

  // loop through the data and send each DWORD
  for (; uByteCnt; uByteCnt -= sizeof(DWORD)) {
    DWORD uCharNum;

    // prepare the event
    RESET_EVENT(&CompletionEvent);

    // Read a DWORD of data from the record
    for (uCharNum = 0; uCharNum < sizeof(DWORD); uCharNum++, pInPtr += 2)
      ArmCommand.CsrParam.Param.WriteMemory.yyyy.uBytes[uCharNum] = atoh(pInPtr, 2);

    Status = ChipALWrite(pThisAdapter,
			 P46_ARM,
			 ADDRESS_ARM_BAR_0,
			 CSR_ARM_CMD,
			 sizeof(ArmCommand), // Length is unused for ARM.
			 (DWORD) &ArmCommand,
			 0);

    // wait on response
    TimeRemain = WAIT_EVENT_TIMEOUT(&CompletionEvent, NsecDelay);
    if (!TimeRemain) {
      DBG_REPORT("Write Data cmd timed out");
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:processIntel32Rec  Write Data cmd timed out"));
      Status = -ETIMEOUT;
      return (Status);
    }
  }

#if 1/*CLN*/
  ArmCommand.CsrParam.Command = 0;
#endif

  return(STATUS_SUCCESS);
}

/******************************************************************************
FUNCTION NAME:
	ARMALP46CompletionRoutine

ABSTRACT:
	Used as	a general purpose completion routine so it can signal an event,
	passed as the Context

RETURN:
	VOID

DETAILS:
******************************************************************************/
LOCAL VOID ARMALP46CompletionRoutine(
    IN PVOID Context) {
  EVENT_HNDL*	Event =	Context;

  // Set the input event, signal Waiting Thread
  SET_EVENT(Event);
}

/****************************************************************************

Name:
	atoh

Description:
	Converts ASCII hex to DWORD

Inputs:
	char *cInStr = pointer to string
	DWORD uNumChars = number of characters in the string

Outputs:
	none
 ***************************************************************************/
LOCAL DWORD atoh(
    IN const char *cInStr,
    DWORD uNumChars) {
  DWORD uSum = 0;

  while (uNumChars--) {
    char ch = *cInStr++;

    if (ch >= '0' &&
	ch <= '9')
      ch -= '0';
    else if (ch >= 'A' &&
	     ch <= 'F')
      ch -= ('A' - 10);
    else if (ch >= 'a' &&
	     ch <= 'f')
      ch -= ('a' - 10);
    else
      continue;

    uSum <<= 4;
    uSum += ch;
  }

  return uSum;
}

/******************************************************************************
FUNCTION NAME:
	CmgmtStartArm

ABSTRACT:
	Start Arm Extended Command Set

RETURN:
	NTSTATUS

DETAILS:
******************************************************************************/
LOCAL NTSTATUS ARMALStartArm(
    IN CDSL_ADAPTER_T	*pThisAdapter) {
  static P46_IO_COMMAND_T	ArmCommand;/*CLN*/
  NTSTATUS		Status;
  int			TimeRemain;
  signed long 		NsecDelay = DOWNLOAD_TIMEOUT;
  //* The Completion and Sync routines are synchronous to allow for
  //* case when Interrupts are not working correctly or the system
  //* is extremely slow.  In such a case we will Timeout and not download
  //* the code.
  EVENT_HNDL			CompletionEvent;
  static volatile ARM_ASYC_IO_T	ArmSync;/*CLN*/

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALStartArm  %p", pThisAdapter));

#if 1/*CLN*/
  if (ArmCommand.CsrParam.Command)
    DBG_REPORT("WoWo");
#endif

  INIT_EVENT(&CompletionEvent);

  do {
    ArmCommand.CsrParam.Command = ARM_GOTO;
    ArmCommand.CsrParam.Param.Goto.Address = JUMP_ADDRESS;
    ArmCommand.CsrParam.NotifyOnComplete = ARMALP46CompletionRoutine;
    ArmCommand.CsrParam.UserContext = &CompletionEvent;

    Status = ChipALWrite(pThisAdapter,
			 P46_ARM,
			 ADDRESS_ARM_BAR_0,
			 CSR_ARM_CMD,
			 sizeof(ArmCommand),
			 (DWORD) &ArmCommand,
			 0); 		// Mask	Value is unused	by ARM
    if (!NT_SUCCESS(Status))
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:ARMALStartArm  ARM GoTo cmd failed, Status=%lx", Status));

    // wait on response
    TimeRemain = WAIT_EVENT_TIMEOUT(&CompletionEvent, NsecDelay);
    if (!TimeRemain) {
      DBG_REPORT("Write Jump Address cmd timed out");
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:ARMALStartArm  Write Timeout waiting on Jump Address"));
      Status = -ETIMEOUT;
      break;
    }
    Status = STATUS_SUCCESS;

    RESET_EVENT(&CompletionEvent);

    // The Response is Interrupt Driven.  We need a sync handler to catch the
    // response from ChipAl
    ArmCommand.CsrParam.NotifyOnComplete = NULL;
    ArmCommand.CsrParam.UserContext = NULL;

    ArmSync.pAdapter = pThisAdapter;
    ArmSync.pEvent = &CompletionEvent;

    Status = ChipALAddEventHandler(pThisAdapter,
				   ARM_FUNCTION,
				   HW_BOOT_READ_MEMORY,
				   HW_DPC_EVENT,
				   ARMSyncRoutine,
				   (void *)&ArmSync);
    ArmSync.Resp2 = 0;

    ArmCommand.CsrParam.Command = ARM_READ_MEMORY;
    ArmCommand.CsrParam.Param.ReadMemory.Address = EXT_COMMAND_ADDRESS;
    ArmCommand.CsrParam.Param.ReadMemory.AccessType = ARM_ACCESS_32;

    Status = ChipALWrite(pThisAdapter,
			 P46_ARM,
			 ADDRESS_ARM_BAR_0,
			 CSR_ARM_CMD,
			 sizeof(ArmCommand),
			 (DWORD) &ArmCommand,
			 0); 		// Mask	Value is unused by ARM
    if (!NT_SUCCESS(Status))
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:ARMALStartArm  ARM read address failed, Status=%lx", Status));

    // wait on response
    TimeRemain = WAIT_EVENT_TIMEOUT(&CompletionEvent, NsecDelay);

    //* Remove Sync Handler
    ChipALRemoveEventHandler(pThisAdapter,
			     ARM_FUNCTION,
			     HW_BOOT_READ_MEMORY,
			     HW_DPC_EVENT,
			     ARMSyncRoutine);

    if(!(TimeRemain && ArmSync.Resp2)) {
      DBG_REPORT("Read Ext cmd timed out");
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:ARMALStartArm  Read Ext Command Address Failed Status(%8lx)", Status));
      Status = -ETIMEOUT;
      break;
    }
    Status = STATUS_SUCCESS;

    RESET_EVENT(&CompletionEvent);

    ArmCommand.CsrParam.Command = ARM_SET_COMMAND_EXTENSION_ROUTINE_ADDRESS;
    ArmCommand.CsrParam.Param.CommandExtAddress.Address = ArmSync.Resp2;
    ArmCommand.CsrParam.NotifyOnComplete = ARMALP46CompletionRoutine;
    ArmCommand.CsrParam.UserContext = &CompletionEvent;
    Status = ChipALWrite(pThisAdapter,
			 P46_ARM,
			 ADDRESS_ARM_BAR_0,
			 CSR_ARM_CMD,
			 sizeof(ArmCommand), // Length is unused for ARM.
			 (DWORD) &ArmCommand,
			 0);		// Mask	Value is unused by ARM
    if (!NT_SUCCESS(Status))
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:ARMALStartArm  ARM set extension cmd failed, Status=%lx", Status));

    // wait on response
    TimeRemain = WAIT_EVENT_TIMEOUT(&CompletionEvent, NsecDelay);
    if (!TimeRemain) {
      DBG_REPORT("Write Ext Address cmd timed out");
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:ARMALStartArm  Timeout waiting on Write of Ext. Address."));
      Status = -ETIMEOUT;
      break;
    }
    Status = STATUS_SUCCESS;
  } while (FALSE);

  if(Status != STATUS_SUCCESS)
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>ARMAbstract:ARMALStartArm  Arm Was Never Ready for Operation!"));

#if 1/*CLN*/
  ArmCommand.CsrParam.Command = 0;
#endif

  return(Status);
}

#define MAX_NUM_GPIO_PINS 	(sizeof(DWORD) * 8)

/****************************************************************************
Name:
	ARMALSetEdge

Description:
	Builds an ARM Set IO Edge command. The ARM
	initializes all GPIO's to trigger on the falling edge.
	This command is used to override the default.

Inputs:
	GPIO pin number.

Outputs:
	GPIO command data structure.
 ***************************************************************************/
LOCAL GPIO_INIT_T ARMALSetEdge(WORD pin) {
  GPIO_INIT_T	GPIO_Init = { { ARM_SET_IO_EDGE, {{0}}, NULL, NULL}, 
			      sizeof(COMMAND_SET_IO_EDGE_T) };

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALSetEdge"));

  //
  //Parameter 1 - GPIO Pin number
  //
  GPIO_Init.CsrParam.Param.IoEdge.Number = pin;

  //
  //Parameter 2 - Edge Trigger, Falling = 1
  //
  GPIO_Init.CsrParam.Param.IoEdge.Trigger = 1;

  return GPIO_Init;
}

/****************************************************************************
Name:
        ARMALSetIoDirection

Description:
	Builds an ARM Set IO Direction command.	The ARM
	initializes all GPIO's as input.
	This command is used to override the default.

Inputs:
	GPIO pin number.

Outputs:
	GPIO command data structure.
 ***************************************************************************/
LOCAL GPIO_INIT_T ARMALSetIoDirection(DWORD GpioDirectionMask, WORD pin) {
  GPIO_INIT_T	GPIO_Init = { { ARM_SET_IO_DIRECTION, {{0}}, NULL, NULL}, 
			      sizeof(COMMAND_SET_IO_DIRECTION_T) };

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALSetIoDirection"));

  //
  //Parameter 1 - GPIO Pin Number
  //
  GPIO_Init.CsrParam.Param.IoDirection.Number = pin;

  //
  //Parameter 2 - Direction
  //
  GPIO_Init.CsrParam.Param.IoDirection.Direction =
                          (GpioDirectionMask >> pin) & 1;

  return GPIO_Init;
}

/****************************************************************************
Name:
	ARMALSetIoAccessPrivileges

Description:
	Builds an ARM Set IO Access Privileges command.

Inputs:
	GPIO pin number.
	GPIO Mask of bits affected.

Outputs:
	GPIO command data structure.
 ***************************************************************************/
LOCAL GPIO_INIT_T ARMALSetIoAccessPrivileges(WORD function, DWORD GpioMask) {
  GPIO_INIT_T	GPIO_Init = { { ARM_SET_IO_ACCESS_PRIV, {{0}}, NULL, NULL}, 
			      sizeof(CSR_COMMAND_PARAM_T) };

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALSetIoAccessPrivileges"));

  //
  //Parameter 1 
  //
  GPIO_Init.CsrParam.Param.IoAccess.Function = function;

  //
  //Parameter 2 
  //
  GPIO_Init.CsrParam.Param.IoAccess.IoAccess = GpioMask;

  return GPIO_Init;	
}

/****************************************************************************
Name:
	ARMALSetIo

Description:
	Builds an ARM Set IO command.	

Inputs:
	GPIO pin number.
	value 1 = ON, 0 = OFF

Outputs:
	GPIO command data structure.
 ***************************************************************************/
LOCAL GPIO_INIT_T ARMALSetIo(WORD pin, WORD value) {
  GPIO_INIT_T	GPIO_Init = { { ARM_SET_IO, {{0}}, NULL, NULL}, 
			      sizeof(DWORD) + sizeof(COMMAND_SET_IO_T)};

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALSetIo"));

  //
  //Parameter 1
  //
  GPIO_Init.CsrParam.Param.SetIo.Number = pin;

  //
  //Parameter 2
  //
  GPIO_Init.CsrParam.Param.SetIo.Value = value;

  return GPIO_Init;
}

/****************************************************************************
Name:
	initIO

Description:
	Initializes GPIO access priviledges and directions

Inputs:
	pThisAdapter = pointer to adapter context

Outputs:
	none
 ***************************************************************************/
NTSTATUS ARMALInitIo(
    IN CDSL_ADAPTER_T	*pThisAdapter) {	
  WORD 			NumCmds = 0;
  WORD 			pin;
  GPIO_INIT_T 		GPIO_Init_Table[MAX_NUM_GPIO_PINS];
  EVENT_HNDL		CompletionEvent;
  NTSTATUS		Status=STATUS_SUCCESS;
  WORD 			Tbl_Index;
  signed long 		NsecDelay = DOWNLOAD_TIMEOUT;
  int			TimeRemain;
  static P46_IO_COMMAND_T	ArmCommand;/*CLN*/

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALInitIo"));

#if 1/*CLN*/
  if (ArmCommand.CsrParam.Command)
    DBG_REPORT("Wo");
#endif

  //
  //Setup ADSL Access Privileges
  //
  GPIO_Init_Table[NumCmds++] = ARMALSetIoAccessPrivileges(1, pThisAdapter->PersistData.GpioADSLFunctionMask);

  //
  //Setup Arbitrator Access Privileges
  //
  GPIO_Init_Table[NumCmds++] = ARMALSetIoAccessPrivileges(0, pThisAdapter->PersistData.GpioARMFunctionMask);

  //
  //Build-up GPIO commands
  //
  for (pin = 0; pin < MAX_NUM_GPIO_PINS; pin++) {
    //
    //Build Set IO Direction commands
    //
    if((pThisAdapter->PersistData.GpioDirectionMask >> pin) & 1) {
      GPIO_Init_Table[NumCmds] = ARMALSetIoDirection(pThisAdapter->PersistData.GpioDirectionMask, pin);
      //Increment 'number of commands' counter
      NumCmds++;
    }

    //
    //Build Set Edge trigger commands
    //
    if((pThisAdapter->PersistData.GpioEdgeMask >> pin) & 1) {
      ARMALSetEdge(pin);
      //Increment 'number of commands' counter
      NumCmds++;
    }	
  }

  //
  //Set initial GPIO bits ON (SSCAN).
  //
  GPIO_Init_Table[NumCmds++] = ARMALSetIo(SSCAN, 1);

  //
  //Set SSCAN I/O Direction
  //
  GPIO_Init_Table[NumCmds] = ARMALSetIoDirection(pThisAdapter->PersistData.GpioDirectionMask, SSCAN);		
  //Notify handler
  GPIO_Init_Table[NumCmds].CsrParam.NotifyOnComplete = ARMALP46CompletionRoutine;
  //Notify context
  GPIO_Init_Table[NumCmds].CsrParam.UserContext = &CompletionEvent;

  NumCmds++;

  INIT_EVENT(&CompletionEvent);

  for (Tbl_Index = 0; Tbl_Index < NumCmds; Tbl_Index++) {
    ArmCommand.CsrParam = GPIO_Init_Table[Tbl_Index].CsrParam;

    Status = ChipALWrite(pThisAdapter,
			 P46_ARM,
			 ADDRESS_ARM_BAR_0,
			 CSR_ARM_CMD,
			 sizeof(P46_IO_COMMAND_T),// Not used at this time
			 (DWORD) &ArmCommand,
			 0); 			// Mask	Value is unused	by CSR_ARM_CMD

    if (!NT_SUCCESS(Status)) {
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>ARMAbstract:ARMALInitIo  Write command failure on write #%d.", Tbl_Index));
      break;
    }
  }

  // wait on response
  TimeRemain = WAIT_EVENT_TIMEOUT(&CompletionEvent, NsecDelay);
  if (!TimeRemain) {
    DBG_REPORT("Write GPIO cmd timed out");
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>ARMAbstract:ARMALInitIo  Write Timout waiting on GPIO."));
    Status = -ETIMEOUT;
  } else {
    ARMALNotifyArmReady(pThisAdapter);
    Status = STATUS_SUCCESS;
  }

#if 1/*CLN*/
  ArmCommand.CsrParam.Command = 0;
#endif

  return Status;
}

/******************************************************************************
FUNCTION NAME:
	ARMALNotifyArmReady

ABSTRACT:
	Notify other functions ARM is now ready to receive commands

	Writes ARM_CONFIGURATION_COMPLETE and then waits on the corresponding
	response.
	This waiting on the response is in order to determine when the last
	ARM response has been received. This is to work around a problem with 
	Qube Linux in that if an interrupt is pending on both functions
	simultaneously, interrupts are no longer serviced.

RETURN:
	VOID

DETAILS:
******************************************************************************/
LOCAL VOID ARMALNotifyArmReady(
    IN CDSL_ADAPTER_T	*pThisAdapter) {
  NTSTATUS			Status;
  static P46_IO_COMMAND_T	ArmCommand;/*CLN*/
  int				TimeRemain;
  EVENT_HNDL			CompletionEvent;
  static volatile ARM_ASYC_IO_T	ArmSync;/*CLN*/

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   (DEV_LABEL "=>ARMAbstract:ARMALNotifyArmReady"));

#if 1/*CLN*/
  if (ArmCommand.CsrParam.Command)
    DBG_REPORT("Wo");
#endif

  // prepare the wait entry
  INIT_EVENT(&CompletionEvent);

  ArmSync.pAdapter = pThisAdapter;
  ArmSync.pEvent = &CompletionEvent;

  Status = ChipALAddEventHandler(pThisAdapter,
				 ARM_FUNCTION,
				 HW_CONFIGURATION_COMPLETE,
				 HW_DPC_EVENT,
				 ARMSyncRoutine,
				 (void *)&ArmSync);

  ArmCommand.CsrParam.Command = ARM_CONFIGURATION_COMPLETE;
  ArmCommand.CsrParam.NotifyOnComplete = NULL;

  Status = ChipALWrite(pThisAdapter,
		       P46_ARM,
		       ADDRESS_ARM_BAR_0,
		       CSR_ARM_CMD,
		       sizeof(ArmCommand),	// unused
		       (DWORD) &ArmCommand,
		       0);		// Mask Value is unused	by ARM

  // wait on response
  TimeRemain = WAIT_EVENT_TIMEOUT(&CompletionEvent, DOWNLOAD_TIMEOUT);
  // Make sure we did not timeout and the ARM sent a response
  if (!TimeRemain) {
    DBG_REPORT("Arm config comp cmd timed out");
    DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		     ChipALDebugFlag,
		     (DEV_LABEL "=>ARMAbstract:ARMALNotifyArmReady  ARM_CONFIGURATION_COMPLETE timed out!"));
    Status = STATUS_FAILURE;
  } else {
    DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		     ChipALDebugFlag,
		     (DEV_LABEL "=>ARMAbstract:ARMALNotifyArmReady  ARM function completed"));
    Status = STATUS_SUCCESS;
  }

#if 1/*CLN*/
  ArmCommand.CsrParam.Command = 0;
#endif

  // write the bit that tells the firmware it is now safe to begin DSL responses
  // (and therefore the first DSL interrupt) since the last ARM interrupt
  // (the ARM_CONFIGURATION_COMPLETE response) has completed.
  if (Status == STATUS_SUCCESS)
    ChipALSetFirmGo(pThisAdapter);
}
