/******************************************************************************
*******************************************************************************
****	Copyright (c) 1998
****	Rockwell Semiconductor Systems
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	ATM
**
**FILE NAME:
**	ATM.h
**
**ABSTRACT:
**
**DETAILS:
**
*******************************************************************************
** $Archive: /Projects/Tigris and Derivatives/Reference Designs/Tigris/Linux/Kernel Driver/ATM.h $
** $Revision: 1 $
** $Date: 1/09/01 10:54a $
*******************************************************************************
******************************************************************************/

#ifndef _ATM_H_
#define _ATM_H_

/* bitfields endian games */

#if defined(__LITTLE_ENDIAN_BITFIELD)
#define BITFIELD2(b1, b2)                    b1; b2;
#define BITFIELD3(b1, b2, b3)                b1; b2; b3;
#define BITFIELD4(b1, b2, b3, b4)            b1; b2; b3; b4;
#define BITFIELD5(b1, b2, b3, b4, b5)        b1; b2; b3; b4; b5;
#define BITFIELD6(b1, b2, b3, b4, b5, b6)    b1; b2; b3; b4; b5; b6;
#elif defined(__BIG_ENDIAN_BITFIELD)
#define BITFIELD2(b1, b2)                                    b2; b1;
#define BITFIELD3(b1, b2, b3)                            b3; b2; b1;
#define BITFIELD4(b1, b2, b3, b4)                    b4; b3; b2; b1;
#define BITFIELD5(b1, b2, b3, b4, b5)            b5; b4; b3; b2; b1;
#define BITFIELD6(b1, b2, b3, b4, b5, b6)    b6; b5; b4; b3; b2; b1;
#else
#error unknown bitfield endianess
#endif

// ATM header Structure

//
// Force misalignment of the following structures
//
// #include <pshpack1.h>

/* This assumes a little endian machine */
typedef struct ATM_HEADER_S {
unsigned int	VPI_UPPER: 4;
unsigned int    GFC: 4;
unsigned int	VCI_UPPER: 4;
unsigned int	VPI_LOWER: 4;
u_int8_t	VCI_MIDDLE;

unsigned int 	CLP: 1;
#if 0
unsigned int	PTI: 3;
#else
unsigned int	PTI_USER: 1;
unsigned int	PTI_CI: 1;
unsigned int	PTI_MGM: 1;
#endif

unsigned int	VCI_LOWER: 4;

u_int8_t	HEC;
} ATM_HEADER_T, *pATM_HEADER_T;

/* The compiler will allocate 8 bytes for an ATM_HEADER_T, so use
   CELL_HEADER_SIZE rather than sizeof() */

#define	CELL_HEADER_SIZE	5

// #include <poppack.h>

#define	CELL_PAYLOAD_SIZE	48

#define SET_VPI(HeaderPtr, VPI) do { \
				HeaderPtr->VPI_UPPER = (VPI >> 4) & 0xf; \
				HeaderPtr->VPI_LOWER = VPI & 0xf; \
				} while (0)

#define GET_VPI(HeaderPtr) ((HeaderPtr->VPI_UPPER << 4) | HeaderPtr->VPI_LOWER)

#define SET_VCI(HeaderPtr, VCI) do { \
				HeaderPtr->VCI_UPPER = (VCI >> 12) & 0xf; \
				HeaderPtr->VCI_MIDDLE = (VCI >> 4) & 0xff; \
				HeaderPtr->VCI_LOWER = VCI & 0xf; \
				} while (0)

#define GET_VCI(HeaderPtr) ((HeaderPtr->VCI_UPPER << 12) | \
			    (HeaderPtr->VCI_MIDDLE << 4) | \
			    (HeaderPtr->VCI_LOWER))

#endif //  _ATM_H_
