/******************************************************************************
*******************************************************************************
****	Copyright (c) 1998, 1999, 2000, 2001
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	FrameAL	ATM layer.
**
**FILE NAME:
**	FrameALATM.c
**
**ABSTRACT:
**	This files contains functions for FrameAL ATM Layer. This includes
**	Traffic Shaping, Mux/DeMux and CellAlign functions.
**
**DETAILS:
**	The CellAlign functions is to fix a problem with the Basic2 Hardware,
**      and is not part of generic ATM.
**
*******************************************************************************
** $Archive: /Projects/Tigris and Derivatives Linux/Reference Designs/Tigris/Linux 2.4 ATM Driver/KernelModule/FrameALATM.c $
** $Revision: 2 $
** $Date: 2/28/01 11:05a $
*******************************************************************************
******************************************************************************/

#include "linuxmodver.h"

#include "CardMgmt.h"
#include "FrameAL.h"
#include "FrameALOAM.h"
#include "FrameALHec.h"
#include "CellDataTestMgmt.h"

// tigris
#define FRAMEALATM_TRAFFIC_SHAPING 1
#define FRAMEALATM_CELL_RCV_INTERFACE 1

#define OAM_VPI 8
#define OAM_VCI 51

#define OAM_VCI_F4_3 3
#define OAM_VCI_F4_4 4

// Cell align states
typedef enum _CELL_ALIGN_STATES {
  // Initialize to start of cell as first data out of the fifo is a cell.
  START_OF_CELL = 0,
  INCOMPLETE_HEADER,
  IN_CELL,
  INCOMPLETE_CELL,
  LOST_ALIGN,
  FLAG_1,
  FLAG_2,
  FLAG_3,
  FLAG_4,
  FLAG_5,
  FLAG_6,
  FLAG_7,
  FLAG_8,
  FLAG_9,
  FLAG_10,
  FLAG_LAST
} CELL_ALIGN_STATES;

typedef enum {
  ATM_LINK_STATE_IDLE = 0,
  ATM_LINK_STATE_ACTIVE
} ATM_LINK_STATE;

typedef enum {
  ATM_RX_CELL_COMPLETE = 0,
  ATM_RX_CELL_PARTIAL
} RX_BUFFER_TYPE;

//*****************************************************************************
//   LOCAL functions prototypes.
//*****************************************************************************
LOCAL  CDSL_VC_T *FrameALATMDeMUX(
    IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
    IN	SK_BUFF_T	*pBufDesc	// Pointer to a buffer descriptor for FrameAL to use
);

LOCAL void FrameALATMGetCell(
    IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
    IN	SK_BUFF_T	*pBufDesc	// Pointer to a buffer descriptor for FrameAL to use
);

/******************************************************************************
FUNCTION NAME:
	FrameALATMLinkInit

ABSTRACT:
	Initializes a new link instance for Mux/Demux.

RETURN:
	NDIS status

DETAILS:
	This function initialize FrameALATM layer
	
******************************************************************************/
GLOBAL NDIS_STATUS FrameALATMLinkInit(
    IN	CDSL_LINK_T	*LinkHdl		// Link Handle to data stream
) {
  FRAMEAL_LINK_T * pFrameALLink = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;
  pATM_HEADER_T pHeader;

  // set link state to idle so start tx calls will be passed thru
  pFrameALLink->ATMLinkBlock.LinkState = ATM_LINK_STATE_IDLE;

  // make sure HEC has generated its encode/decode table
  FrameALHECInit();

  // Initailize the Traffic Shaper
#if FRAMEALATM_TRAFFIC_SHAPING
  FrameATMShLinkInit(LinkHdl);
#endif

#if HAS_IDLEHEADER
  // init idle cell
  pHeader = &pFrameALLink->ATMLinkBlock.IdleHeader;

  CLEAR_MEMORY(pHeader, CELL_HEADER_SIZE);

  pHeader->CLP = 1;

  FrameALHecEncode((BYTE *) pHeader);
#endif

#if HAS_IDLEPAYLOAD
  MEMORY_SET(&pFrameALLink->ATMLinkBlock.IdlePayload, IDLE_PAYLOAD_CHAR,
	     CELL_PAYLOAD_SIZE);
#endif

#if HAS_LINKOAM
  // init OAM cell
  pHeader = &pFrameALLink->ATMLinkBlock.LinkOAMHeader;

  pHeader->GFC = 0;

  SET_VPI(pHeader, OAM_VPI);
  SET_VCI(pHeader, OAM_VCI);

  pHeader->CLP = 1;

  pHeader->PTI_USER = 1;
  pHeader->PTI_CI = 0;
  pHeader->PTI_MGM = 1;

  // calculate the HEC for this cell header
  FrameALHecEncode((BYTE *) pHeader);
#endif

  // setup for OAM for this connection
  FrameALATMOAMLinkInit(LinkHdl);

  return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	FrameALATMVCInit

ABSTRACT:
	Initializes a new VC instance.

RETURN:
	NDIS status

DETAILS:
	This function initializes VC instance which includes ATM cell headers.

******************************************************************************/
GLOBAL NDIS_STATUS FrameALATMVCInit(
    IN	CDSL_VC_T	*VCHdl			// Link Handle to data stream
) {
  if (VCHdl) {
    FRAMEAL_VC_T  		*pFrameALVC = (FRAMEAL_VC_T*)VCHdl->FrameALVCMod;
    WORD 			LoopCounter;
    FRAMEAL_SHAPER_STATUS	OldTrafficShapingFlag;
    FRAMEAL_LINK_T 		*pFrameALLink;

#if FRAMEALATM_TRAFFIC_SHAPING
    if (VCHdl->LinkHdl && VCHdl->LinkHdl->FrameMod) {
      // Update traffic shaper status
      pFrameALLink = (FRAMEAL_LINK_T*) VCHdl->LinkHdl->FrameMod;

      // turn off traffic shaping while updating the ATM layer
      OldTrafficShapingFlag = pFrameALLink->ATMLinkBlock.TrafficShapingActive;
      pFrameALLink->ATMLinkBlock.TrafficShapingActive = TRAFFIC_SHAPING_OFF;
    } else
#endif
    {
      pFrameALLink = NULL;
      OldTrafficShapingFlag = FALSE;
    }
    //
    // Fill in all possible Cell Headers
    //
    for (LoopCounter = 0;LoopCounter < PRECOMPUTED_HEADER_NUM;LoopCounter++) {
#if PRECOMPUTED_HEADERS
      pATM_HEADER_T 	pHeader = &pFrameALVC->ATMVCBlock.TxPreCompHeader[LoopCounter];
#else
      pATM_HEADER_T 	pHeader = &pFrameALVC->ATMVCBlock.TxPreCompHeader;
#endif
      OOB_DATA_T oob_data;

      oob_data.v = LoopCounter;

      pHeader->GFC = oob_data.b.GFC;

      // fill in VCI and VPI fields of the header
      SET_VCI(pHeader, pFrameALVC->Parameters.Vci);
      SET_VPI(pHeader, pFrameALVC->Parameters.Vpi);

      pHeader->CLP = oob_data.b.CLP;

      pHeader->PTI_USER = oob_data.b.PTI_USER;
      pHeader->PTI_CI = oob_data.b.PTI_CI;
      pHeader->PTI_MGM = oob_data.b.PTI_MGM;

      // calculate the HEC for this cell header
      FrameALHecEncode((BYTE *) pHeader);

#if PRECOMPUTED_HEADERS
#else
      pFrameALVC->ATMVCBlock.TxPreCompHEC[LoopCounter] =
	pFrameALVC->ATMVCBlock.TxPreCompHeader.HEC;
#endif
    }

    // fill in SAP service routine addresses based on SAP info
    switch (pFrameALVC->Parameters.VcSAP) {
    case VC_SAP_AAL:
      // set the addresses of AAL's server routines into the VC entry
      pFrameALVC->ATMVCBlock.TxGetFullBuf  = FrameALAALTxGetFullBuf;
      pFrameALVC->ATMVCBlock.TxPutEmptyBuf = FrameALAALTxPutEmptyBuf;
      pFrameALVC->ATMVCBlock.RxPutFullBuf  = FrameALAALRxPutFullBuf;
      break;

    case VC_SAP_CELLDATATEST:
      // set the addresses of AAL's server routines into the VC entry
      pFrameALVC->ATMVCBlock.TxGetFullBuf  = CellDataTestMgmtTxGetFullBuf;
      pFrameALVC->ATMVCBlock.TxPutEmptyBuf = CellDataTestMgmtTxPutEmptyBuf;
      pFrameALVC->ATMVCBlock.RxPutFullBuf  = CellDataTestMgmtRxPutFullBuf;
      break;

    default:
      pFrameALLink->ATMLinkBlock.TrafficShapingActive = OldTrafficShapingFlag;
      return STATUS_FAILURE;
    }

#if FRAMEALATM_TRAFFIC_SHAPING
    if (VCHdl->LinkHdl && VCHdl->LinkHdl->FrameMod) {
      FRAMEAL_SHAPER_STATUS	TempTrafficShapingFlag;

      // Add this VC to the traffic shaper
      TempTrafficShapingFlag = FrameATMShVCInit(VCHdl);
      pFrameALLink->ATMLinkBlock.TrafficShapingActive = 
	(TempTrafficShapingFlag == TRAFFIC_SHAPING_NOCHANGE) ?
	OldTrafficShapingFlag : TempTrafficShapingFlag;
    }
#endif
    // set up OAM for this VC
    FrameALATMOAMVCInit(VCHdl->LinkHdl, VCHdl);

    return STATUS_SUCCESS;
  } else
    return STATUS_FAILURE;
}

/******************************************************************************
FUNCTION NAME:
	FrameALATMLinkshutdown

ABSTRACT:
	Shuts down an ATM layer's link instance.

RETURN:
	NDIS Status

DETAILS:
	The memory for the FrameAL ATM layer link context is released.
	Also, any cells in the loopback queue are released.

******************************************************************************/
GLOBAL NDIS_STATUS FrameALATMLinkShutdown(
    IN	CDSL_LINK_T	*LinkHdl		// Link Handle to data stream
) {
  // shutdown Link traffic shaper
  FrameATMShLinkShutdown(LinkHdl);

  // shutdown Link OAM
  FrameALATMOAMLinkShutdown(LinkHdl);
	
  return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	FrameALATMVChutdown

ABSTRACT:
	Shuts down an ATM layer's VC instance.

RETURN:
	NDIS Status

DETAILS:
	The memory for the FrameAL ATM layer link context is released.

******************************************************************************/
GLOBAL NDIS_STATUS FrameALATMVCShutdown(
    IN	CDSL_VC_T	*VCHdl			// Link Handle to data stream
) {
  if (VCHdl->LinkHdl && VCHdl->LinkHdl->FrameMod) {
    FRAMEAL_VC_T	*pFrameALVC = (FRAMEAL_VC_T *) VCHdl->FrameALVCMod;

    //turn off data for this VC
    if (pFrameALVC) {
      DWORD 		Index;

      pFrameALVC->ATMVCBlock.TxGetFullBuf = NULL;
      pFrameALVC->ATMVCBlock.RxPutFullBuf = NULL;

      // check for any calls to upper layers still active
      for (Index = 0; Index < FRAMEALVC_CNXT_NO; Index++)
	if (pFrameALVC->ATMVCBlock.InVCALCNXT[Index])
	  return COMMAND_PENDING;
    }

#if FRAMEALATM_TRAFFIC_SHAPING
    {
      FRAMEAL_LINK_T 	*pFrameALLink = (FRAMEAL_LINK_T*) VCHdl->LinkHdl->FrameMod;
      FRAMEAL_SHAPER_STATUS	OldTrafficShapingFlag;
      FRAMEAL_SHAPER_STATUS	TempTrafficShapingFlag;

      // turn off traffic shaping while updating the ATM layer
      OldTrafficShapingFlag = pFrameALLink->ATMLinkBlock.TrafficShapingActive;
      pFrameALLink->ATMLinkBlock.TrafficShapingActive = TRAFFIC_SHAPING_OFF;

      // remove this VC from the traffic shaper
      TempTrafficShapingFlag = FrameATMShVCShutdown(VCHdl);
      pFrameALLink->ATMLinkBlock.TrafficShapingActive = 
	(TempTrafficShapingFlag == TRAFFIC_SHAPING_NOCHANGE) ?
	OldTrafficShapingFlag : TempTrafficShapingFlag;
    }
#endif 	

    FrameALATMOAMVCShutdown(VCHdl);
  }

  return STATUS_SUCCESS;
}

LOCAL void DumpCell(
 IN	const char	*m,
 IN	const SK_BUFF_T	*pBufDesc
) {
  const pATM_HEADER_T 	pCellHeader = CELLHEAD_DATA(pBufDesc);

  printk(KERN_INFO "CnxADSL: %s Vpi:%d Vci:%d "
	 "GFC:%d PTI_MGM:%d PTI_CI:%d PTI_USER:%d CLP:%d HEC:%0x\n",
	 m,
	 GET_VPI(pCellHeader), GET_VCI(pCellHeader),
	 pCellHeader->GFC,
	 pCellHeader->PTI_MGM,
	 pCellHeader->PTI_CI,
	 pCellHeader->PTI_USER,
	 pCellHeader->CLP,
	 pCellHeader->HEC);
  if (0)
    if (pBufDesc->data && pBufDesc->len) {
      const u_int8_t *b = pBufDesc->data;
      size_t i;
      char line[3*16+1];

      for (i = 0; i < pBufDesc->len; i++, b++) {
	sprintf(line+(i & 0xf)*3, "%02x ", *b);
	if ((i & 0xf) == 0xf)
	  printk(KERN_INFO "CnxADSL: %s %s\n",
		 m,
		 line);
      }
    }
}

/******************************************************************************
FUNCTION NAME:
	FrameALMUXGetTxFullBuffer

ABSTRACT:
	Get a cell from one of the active VCs.

RETURN:
	Pointer to NDIS WAN Packet descriptor

DETAILS:
	This function gets a buffer containing one ATM cell from SAR.
	
******************************************************************************/

GLOBAL SK_BUFF_T	*FrameALATMGetTxFullBuffer(
 IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
 IN	SK_BUFF_T	*pBufDesc	// Pointer to a buffer descriptor for FrameAL to use
) {
  if (LinkHdl) {
    FRAMEAL_LINK_T  *pFrameALLink = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;

    if (pFrameALLink) {
      // call sar interface function this link
      FrameALATMGetCell(LinkHdl, pBufDesc);

      if (pBufDesc->len > 0) {
	if (!OOB_DATA(pBufDesc).b.IDLE) {
	  pATM_HEADER_T  pATMHeader = CELLHEAD_DATA(pBufDesc);

	  ADD_TO_CTR(pFrameALLink->Stats.TXATMCellsCLP[pATMHeader->CLP], 1);
	  ADD_TO_CTR(pFrameALLink->Stats.TXFrames, 1);
	  ADD_TO_CTR(pFrameALLink->Stats.TXBytes,
		     CELLHEAD_LEN(pBufDesc)+pBufDesc->len);

	  ADD_TO_CTR(pFrameALLink->Stats.COTXATMCells, 1);
	  ADD_TO_CTR(pFrameALLink->Stats.COTXBytes,
		     CELLHEAD_LEN(pBufDesc)+pBufDesc->len);
	}
	pFrameALLink->ATMLinkBlock.LinkState = ATM_LINK_STATE_ACTIVE;
      } else // set link state to idle so start tx calls will be passed thru
	pFrameALLink->ATMLinkBlock.LinkState = ATM_LINK_STATE_IDLE;
    }
  }

  return pBufDesc;
}

/******************************************************************************
FUNCTION NAME:
	FrameALMUXTxPutEmptyBuf

ABSTRACT:
	Handles buffer for a cell that has just been transmitted.

RETURN:
	Pointer to NDIS WAN Packet descriptor

DETAILS:
	This function checks to see if the buffer is for a loopback cell or a
	cell from the AAL layer. If the buffer is for a loopback cell, release 
	the memory used for the buffer. If the buffer is for an AAL layer cell
	call AAL layer and let it update its pointers.

******************************************************************************/
GLOBAL SK_BUFF_T	*FrameALATMTxPutEmptyBuf(
    IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
    IN	SK_BUFF_T	*pBufDesc	// Pointer to a buffer descriptor for FrameAL to use
) {
  if (pBufDesc &&
      pBufDesc->len > 0 &&
      !OOB_DATA(pBufDesc).b.IDLE &&
      !OOB_DATA(pBufDesc).b.OAM_CELL) {
    CDSL_VC_T			* pVChdl = VC_HNDL_DATA(pBufDesc);

    if (pVChdl) {
      FRAMEAL_VC_T 		* pFrameALVC;

      pFrameALVC = (FRAMEAL_VC_T *)pVChdl->FrameALVCMod;
      if (pFrameALVC &&
	  pFrameALVC->ATMVCBlock.TxPutEmptyBuf) {
	pFrameALVC->ATMVCBlock.InVCALCNXT[FRAMEALVC_CNXT_PUTTX] = TRUE;
	(* pFrameALVC->ATMVCBlock.TxPutEmptyBuf)(VC_HNDL_DATA(pBufDesc), pBufDesc);
	pFrameALVC->ATMVCBlock.InVCALCNXT[FRAMEALVC_CNXT_PUTTX] = FALSE;
      }
    }
  }

  return pBufDesc;
}

/******************************************************************************
FUNCTION NAME:
	FrameALATMCellAlign

ABSTRACT:
	Search the receive buffer for good cells.

RETURN:
	Pointer to NDIS WAN Packet descriptor

DETAILS:
	This function processes raw receive buffers. The function extracts
	cells from the receive buffer. Cells are identified by a header with
	a valid HEC and headers are preceeded by the 11 byte file pattern
	(except for the first cell in the buffer). Since the actual buffer is
	a circular buffer ChipAl handles wraps by splitting the buffer in two
	pieces (first piece last position to end of buffer and second piece
	start of buffer to current position). Since Chipal passes two pieces,
	this function must consume all data in the buffer before returning it
	to ChipAl.

	When all the data in the buffer is consumed, this function will notify
	AAL and Bufmgmt of end of buffer. They will call NDIS complete
	functions.

******************************************************************************/
GLOBAL SK_BUFF_T	*FrameALATMCellAlign(
 IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
 IN	SK_BUFF_T	*BufferDescPtr	// Pointer to a buffer descriptor
) {
  if (LinkHdl) {
#if FRAMEALATM_CELL_RCV_INTERFACE
    // clear upper bits used by frameAL, then set OOD_END_CELL
    OOB_DATA(BufferDescPtr).v = 0;
    OOB_DATA(BufferDescPtr).b.START_CELL = 1;
    OOB_DATA(BufferDescPtr).b.END_CELL = 1;

    CELLHEAD_LEN(BufferDescPtr) &= 0xffff;

    {
      CDSL_VC_T 	  * pCellVC;

      pCellVC = FrameALATMDeMUX(LinkHdl,BufferDescPtr);
#if HAS_CELLSTOSAP
      if (pCellVC) {
	FRAMEAL_VC_T	* pFrameALVC = (FRAMEAL_VC_T *) pCellVC->FrameALVCMod;

	if (pFrameALVC)
	  pFrameALVC->ATMVCBlock.CellsToSap = 0;
      }
#endif
    }
#else
    {
      FRAMEAL_LINK_T  * pFrameALLink = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;
      BYTE			*pBufferPointer;
      SK_BUFF_T			TempDesc;
      DWORD			CurrentIndex;
      BOOLEAN			SearchLoopControl;
      DWORD			NewCellCount;
      DWORD			NewPayloadCount;
      DWORD			TempIdleCell;
      DWORD			ReadStatus;

      //
      // Get the Idle character from the hardware via ChipAL hal
      //
      ReadStatus = ChipALRead(LinkHdl->pThisAdapter, // adapter handle
			      HOBBES_FPGA, 		// Hardware device
			      HOBBES_RX_IDLE_CHAR,
			      0, /*???*/
			      sizeof(BYTE), 	// size of data to be read
			      &TempIdleCell);
      pFrameALLink->ATMLinkBlock.IdleCharacter = (BYTE) TempIdleCell;

      if (ReadStatus != CHIPAL_SUCCESS)
	;

      // if a HARDWARE loopback is needed,
      // set the loopback bit in the Hobbs control register
#if 0
      ReadStatus = ChipALWrite(LinkHdl->pThisAdapter, // adapter handle
			       HOBBES_FPGA, 	// Hardware device
			       HOBBES_CONTROL,
			       sizeof(BYTE), 	// size of data to be read
			       LOOP1, 		// value to be written
			       LOOP1);		// mask value
      if (ReadStatus != CHIPAL_SUCCESS)
	;
#endif

      // Set up the loop to process the buffer of raw data
      CurrentIndex = 0;
      pBufferPointer = BufferDescPtr->data;

      // Loop thru the raw receive buffer finding and processing cells
      SearchLoopControl = TRUE;
      do {
	switch (pFrameALLink->ATMLinkBlock.CellAlignState) {
	case START_OF_CELL:
	  // first check for a complete cell in the buffer
	  if ((BufferDescPtr->len - CurrentIndex) <
	      (CELL_HEADER_SIZE + CELL_PAYLOAD_SIZE)) {
	    // reset payload count
	    pFrameALLink->ATMLinkBlock.RxPayloadCount = 0;

	    // First, check that the complete header is in the buffer
	    if ((BufferDescPtr->len - CurrentIndex) < CELL_HEADER_SIZE) {
	      // copy first portion of the cell to the tempory header buffer
	      // in the the link structure
	      memcpy(&pFrameALLink->ATMLinkBlock.RxCellHeader, pBufferPointer,
		     BufferDescPtr->len - CurrentIndex);

	      pFrameALLink->ATMLinkBlock.RxHeaderCount =
		BufferDescPtr->len - CurrentIndex;

	      CurrentIndex = BufferDescPtr->len;

	      pFrameALLink->ATMLinkBlock.CellAlignState = INCOMPLETE_CELL;
	      SearchLoopControl = FALSE;
	    } else {
	      // check HEC, if good goto in cell state and process the cell
	      // NOTE: the incell state could be moved to this
	      // if statement block

	      if (FrameALHecDecode(pBufferPointer, FALSE) == HEC_GOOD) {
		// the link structure
		memcpy(&pFrameALLink->ATMLinkBlock.RxCellHeader, pBufferPointer,
		       CELL_HEADER_SIZE);

		pFrameALLink->ATMLinkBlock.RxHeaderCount = CELL_HEADER_SIZE;

		// set header pointer in wan packet descriptor
		// to point to cell header
		CELLHEAD_DATA(&TempDesc) = (pATM_HEADER_T)pBufferPointer;
		CELLHEAD_LEN(&TempDesc) = CELL_HEADER_SIZE;

		// move buffer pointer to cell payload
		pBufferPointer += CELL_HEADER_SIZE;
		CurrentIndex += CELL_HEADER_SIZE;

		// set current buffer pointer of WAN packet descriptor to cell payload
		TempDesc.data = pBufferPointer;
		TempDesc.len = BufferDescPtr->len - CurrentIndex;

		// clear bits used by frameAL
		OOB_DATA(&TempDesc).v = 0;

		// move buffer pointer past cell payload
		pBufferPointer += BufferDescPtr->len - CurrentIndex;
		CurrentIndex += TempDesc.len;

		// send the cell to demux which will send the cell to SAR
		pFrameALLink->ATMLinkBlock.RxPayloadCount = TempDesc.len;

		FrameALATMDeMUX(LinkHdl,&TempDesc);

		CurrentIndex = BufferDescPtr->len;

		pFrameALLink->ATMLinkBlock.CellAlignState = INCOMPLETE_CELL;
		SearchLoopControl = FALSE;
		break;
	      } else {
		// Not a good HEC, drop in to search mode
	      NotGoodHEC:
		pFrameALLink->ATMLinkBlock.CellAlignState =
		  (*pBufferPointer == pFrameALLink->ATMLinkBlock.IdleCharacter) ?
		  FLAG_1 : LOST_ALIGN;

		pBufferPointer++;
		CurrentIndex++;
	      }
	    }
	  } else {
	    // check HEC, if good goto in cell state and process the cell
	    // NOTE: the incell state could be moved to this
	    // if statement block

	    if (FrameALHecDecode(pBufferPointer, FALSE) == HEC_GOOD)
	      pFrameALLink->ATMLinkBlock.CellAlignState = IN_CELL;
	    else {
	      // Not a good HEC, drop in to search mode
	      goto NotGoodHEC;

	      pFrameALLink->ATMLinkBlock.CellAlignState =
		(*pBufferPointer == pFrameALLink->ATMLinkBlock.IdleCharacter) ?
		FLAG_1 : LOST_ALIGN;

	      pBufferPointer++;
	      CurrentIndex++;
	    }
	  }
	  break;

	case INCOMPLETE_CELL:
	  NewCellCount = 0;

	  // first check for a complete header in temp header buffer
	  if (pFrameALLink->ATMLinkBlock.RxHeaderCount < CELL_HEADER_SIZE) {
	    NewCellCount = CELL_HEADER_SIZE - pFrameALLink->ATMLinkBlock.RxHeaderCount;

	    if ((BufferDescPtr->len - CurrentIndex) < NewCellCount)
	      NewCellCount = BufferDescPtr->len - CurrentIndex;

	    // copy first portion of the cell to the tempory header buffer in the
	    // the link structure
	    memcpy(((BYTE*)&pFrameALLink->ATMLinkBlock.RxCellHeader) +
		    pFrameALLink->ATMLinkBlock.RxHeaderCount, pBufferPointer,
		   NewCellCount);

	    pBufferPointer += NewCellCount;
	    CurrentIndex += NewCellCount;

	    if ((pFrameALLink->ATMLinkBlock.RxHeaderCount + NewCellCount) <
		CELL_HEADER_SIZE) {
	      pFrameALLink->ATMLinkBlock.RxHeaderCount += NewCellCount;
	      SearchLoopControl = FALSE;
	      // check if now have a complete cell
	      break;
	    }

	    // Determine if the HEC needs to tested.
	    if ((pFrameALLink->ATMLinkBlock.RxHeaderCount + NewCellCount) <
		CELL_HEADER_SIZE) {
	      if (FrameALHecDecode((void *)&pFrameALLink->ATMLinkBlock.RxCellHeader, FALSE) != HEC_GOOD) {
		pFrameALLink->ATMLinkBlock.CellAlignState =
		  (*pBufferPointer == pFrameALLink->ATMLinkBlock.IdleCharacter) ?
		  FLAG_1 : LOST_ALIGN;

		pBufferPointer++;
		CurrentIndex++;
		break;
	      }
	    }
	  }

	  // set header pointer in wan packet descriptor to point to cell header
	  CELLHEAD_DATA(&TempDesc) = &pFrameALLink->ATMLinkBlock.RxCellHeader;
	  CELLHEAD_LEN(&TempDesc) = CELL_HEADER_SIZE;

	  // set current buffer pointer of WAN packet descriptor to cell payload
	  TempDesc.data = pBufferPointer;
	  NewPayloadCount = CELL_PAYLOAD_SIZE - pFrameALLink->ATMLinkBlock.RxPayloadCount;

	  if (NewPayloadCount > (BufferDescPtr->len - CurrentIndex))
	    NewPayloadCount = BufferDescPtr->len - CurrentIndex;

	  TempDesc.len = NewPayloadCount;

	  // move buffer pointer past cell payload
	  pBufferPointer += NewPayloadCount;
	  CurrentIndex += NewPayloadCount;

	  // clear bits used by frameAL
	  OOB_DATA(&TempDesc).v = 0;

	  if ((NewPayloadCount + pFrameALLink->ATMLinkBlock.RxPayloadCount) <
	      CELL_PAYLOAD_SIZE) {
	    pFrameALLink->ATMLinkBlock.CellAlignState = INCOMPLETE_CELL;
	    SearchLoopControl = FALSE;
	    pFrameALLink->ATMLinkBlock.RxPayloadCount += NewPayloadCount;
	    OOB_DATA(&TempDesc).b.START_CELL = 1;
	  } else {
	    pFrameALLink->ATMLinkBlock.CellAlignState = FLAG_1;
	    OOB_DATA(&TempDesc).b.END_CELL = 1;
	  }

	  // send the cell to demux which will send the cell on to SAR
	  FrameALATMDeMUX(LinkHdl,&TempDesc);
	  break;

	case IN_CELL:
	  // Found a cell (good HEC), first check for a complete cell in the buffer
	  if ((BufferDescPtr->len - CurrentIndex) <
	      (CELL_HEADER_SIZE + CELL_PAYLOAD_SIZE)) {
	    // complete cell is not in the buffer, exit and wait for next buffer
	    SearchLoopControl = FALSE;
	    break;
	  }

	  // set WAN packet descriptor to point to cell header
	  CELLHEAD_DATA(&TempDesc) = (pATM_HEADER_T)pBufferPointer;
	  CELLHEAD_LEN(&TempDesc) = CELL_HEADER_SIZE;

	  // move buffer pointer to cell payload
	  pBufferPointer += CELL_HEADER_SIZE;
	  CurrentIndex += CELL_HEADER_SIZE;

	  // set WAN packet descriptor to cell payload
	  TempDesc.data = pBufferPointer;
	  TempDesc.len = CELL_PAYLOAD_SIZE;

	  // move buffer pointer past cell payload
	  pBufferPointer += CELL_PAYLOAD_SIZE;
	  CurrentIndex += CELL_PAYLOAD_SIZE;

	  pFrameALLink->ATMLinkBlock.CellAlignState = FLAG_1;
	  // send the cell to demux which will send the cell on to SAR

	  // mark this cell as complete
	  OOB_DATA(&TempDesc).v = 0;
	  OOB_DATA(&TempDesc).b.START_CELL = 1;
	  OOB_DATA(&TempDesc).b.END_CELL = 1;

	  FrameALATMDeMUX(LinkHdl,&TempDesc);
	  break;

	case LOST_ALIGN:
	  // search for first idle character
	  if (*pBufferPointer == pFrameALLink->ATMLinkBlock.IdleCharacter) {
	    pFrameALLink->ATMLinkBlock.CellAlignState = FLAG_1;

	    // in ATM, use rx error counter to report lost align errors
	    ADD_TO_CTR(pFrameALLink->Stats.RXATMCellsBadHEC, 1);
	    ADD_TO_CTR(pFrameALLink->Stats.CORXATMCellsDrops, 1);
	  }

	  pBufferPointer++;
	  CurrentIndex++;
	  break;

	case FLAG_1:
	case FLAG_2:
	case FLAG_3:
	case FLAG_4:
	case FLAG_5:
	case FLAG_6:
	case FLAG_7:
	case FLAG_8:
	case FLAG_9:
	case FLAG_10:
	  // Check for first idle character
	  if (*pBufferPointer == pFrameALLink->ATMLinkBlock.IdleCharacter) {
	    pFrameALLink->ATMLinkBlock.CellAlignState++;
	    pBufferPointer++;
	    CurrentIndex++;
	  } else
	    pFrameALLink->ATMLinkBlock.CellAlignState = START_OF_CELL;
	  break;

	case FLAG_LAST:
	  // Check for idle character
	  // If it is idle character, it should be the last one before a header
	  if (*pBufferPointer == pFrameALLink->ATMLinkBlock.IdleCharacter) {
	    pBufferPointer++;
	    CurrentIndex++;
	  }

	  pFrameALLink->ATMLinkBlock.CellAlignState = START_OF_CELL;
	  break;

	default:
	  // should never get a default case
	  ASSERT(FALSE);
	  break;
	}

	// Check for end of buffer
	if (CurrentIndex == BufferDescPtr->len)
	  SearchLoopControl = FALSE;
      } while (SearchLoopControl);

      // Inform ChipAL how much of the buffer was consumed
      BufferDescPtr->len = CurrentIndex;

#if HAS_CELLSTOSAP
      {
	DWORD				VCIndex;

	// Determine if AAL Rx complete needs to called
	for (VCIndex = 0; VCIndex < MAX_VC_PER_LINK; VCIndex++) {
	  CDSL_VC_T		*pVCHdl;

	  pVCHdl = &LinkHdl->VC_Info[VCIndex];
	  if (pVCHdl) {
	    FRAMEAL_VC_T	  * pFrameALVC;

	    pFrameALVC = (FRAMEAL_VC_T *) pVCHdl->FrameALVCMod;
	    if (pFrameALVC)
	      pFrameALVC->ATMVCBlock.CellsToSap = 0;
	  }
	}
      }
#endif
    }
#endif
  }

  // Return the WAN packet descriptor to ChipAL
  return BufferDescPtr;
}
/******************************************************************************
FUNCTION NAME:
	FrameALATMDeMux

ABSTRACT:
	Send a cell to SAR with the proper VC handle.

RETURN:
	None

DETAILS:
	This function sends a buffer containing one ATM cell from SAR.
	The function just calls SAR with the link handle.

******************************************************************************/

LOCAL CDSL_VC_T	*FrameALATMDeMUX(
    IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
    IN	SK_BUFF_T	*pBufDesc	// Pointer to a buffer descriptor for FrameAL to use
) {
  CDSL_VC_T		*pVCHdl = NULL;

  if (LinkHdl) {
    pATM_HEADER_T 	pCellHeader = CELLHEAD_DATA(pBufDesc);
    WORD 		TempVci;
    WORD 		TempVpi;

    TempVci = GET_VCI(pCellHeader);
    TempVpi = GET_VPI(pCellHeader);

    // check not idle cell
    if (!((TempVci == 0) && (TempVpi == 0))) {
      FRAMEAL_LINK_T 	*pFrameALLink = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;
      WORD		VCIndex;
      FRAMEAL_VC_T  	*pFrameALVC = NULL;
      BOOLEAN		hasVC = FALSE;

      for (VCIndex = 0; VCIndex < MAX_VC_PER_LINK; VCIndex++) {
	CDSL_VC_T		*pVCHdl_x = &LinkHdl->VC_Info[VCIndex];

	if (pVCHdl_x) {
	  pFrameALVC = (FRAMEAL_VC_T *) pVCHdl_x->FrameALVCMod;
	  if (pFrameALVC) {
	    hasVC = TRUE;
	    // check VCI/VPI for correct match
	    if ((pFrameALVC->Parameters.Vci == TempVci) &&
		(pFrameALVC->Parameters.Vpi == TempVpi)) {
	      if (OOB_DATA(pBufDesc).b.END_CELL) {
		ADD_TO_CTR(pFrameALLink->Stats.RXATMCellsGoodCLP[pCellHeader->CLP], 1);
		ADD_TO_CTR(pFrameALVC->Stats.RXATMCellsGoodCLP[pCellHeader->CLP], 1);
		ADD_TO_CTR(pFrameALLink->Stats.CORXATMCells, 1);
		ADD_TO_CTR(pFrameALVC->Stats.CORXATMCells, 1);

		ADD_TO_CTR(pFrameALLink->Stats.RXFrames, 1);

		ADD_TO_CTR(pFrameALLink->Stats.RXBytes, CELL_HEADER_SIZE + CELL_PAYLOAD_SIZE);
		ADD_TO_CTR(pFrameALVC->Stats.RXBytes, CELL_HEADER_SIZE + CELL_PAYLOAD_SIZE);

		ADD_TO_CTR(pFrameALLink->Stats.CORXBytes, CELL_HEADER_SIZE + CELL_PAYLOAD_SIZE);
		ADD_TO_CTR(pFrameALVC->Stats.CORXBytes, CELL_HEADER_SIZE + CELL_PAYLOAD_SIZE);
	      }
	      pVCHdl = pVCHdl_x;
	      break;
	    }
	  }
	}
      }

      OOB_DATA(pBufDesc).b.CLP = pCellHeader->CLP;
      OOB_DATA(pBufDesc).b.PTI_USER = pCellHeader->PTI_USER;
      OOB_DATA(pBufDesc).b.PTI_CI   = pCellHeader->PTI_CI;
      OOB_DATA(pBufDesc).b.PTI_MGM  = pCellHeader->PTI_MGM;

#if 0 /* not for linux */
      // check for mgmt cells
      if ((TempVci == OAM_VCI_F4_3) || (TempVci == OAM_VCI_F4_4)) {
	FrameALOAMRxBuffer(LinkHdl,pVCHdl,pBufDesc);

	pVCHdl = NULL;
      } else
#endif
	if (pVCHdl) {
	  if (pCellHeader->PTI_MGM) {
	    FrameALOAMRxBuffer(LinkHdl,pVCHdl,pBufDesc);

	    pVCHdl = NULL;
	  } else {
#if HAS_CELLSTOSAP
	    // mark this vc that it delivered a cell to aal.
	    pFrameALVC->ATMVCBlock.CellsToSap++;
#endif

	    if (pFrameALVC->ATMVCBlock.RxPutFullBuf) {
	      pFrameALVC->ATMVCBlock.InVCALCNXT[FRAMEALVC_CNXT_PUTRX] = TRUE;
	      (*pFrameALVC->ATMVCBlock.RxPutFullBuf)(pVCHdl, pBufDesc);
	      pFrameALVC->ATMVCBlock.InVCALCNXT[FRAMEALVC_CNXT_PUTRX] = FALSE;
	    }
	  }
	} else {
	  if (OOB_DATA(pBufDesc).b.END_CELL) {
	    if (!hasVC) /* Nothing in use */
	      DumpCell("Misrouted", pBufDesc);

	    ADD_TO_CTR(pFrameALLink->Stats.RXATMCellsMisrouted, 1);
	    ADD_TO_CTR(pFrameALLink->Stats.CORXATMCellsDrops, 1);
	  }
	}
    }
  }

  // call SAR with the cell if VC handle (Link Handle) non NULL
  return pVCHdl;
}

/******************************************************************************
FUNCTION NAME:
	FrameALATMGetCell

ABSTRACT:
	Get a cell from SAR with the proper VC handle.

RETURN:
	None

DETAILS:
	This function gets a buffer containing one ATM cell from SAR.
        The function just calls SAR with the link handle.

******************************************************************************/

LOCAL void FrameALATMGetCell(
    IN	CDSL_LINK_T		*LinkHdl,	// Link Handle to data stream
    IN	SK_BUFF_T		*pBufDesc	// Pointer to a buffer descriptor for FrameAL to use
) {
  // make sure buffer is marked as empty
  pBufDesc->len = 0;

  // clear upper bits used by frameAL
  OOB_DATA(pBufDesc).v = 0;

  if (LinkHdl) {
    FRAMEAL_LINK_T  	*pFrameALLink;
    FRAMEAL_VC_T	*pFrameALVC;

    pFrameALLink = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;

    FrameALOAMLinkGetTxFullBuffer(LinkHdl,pBufDesc);
    if (pBufDesc->len) {
      if (VC_HNDL_DATA(pBufDesc)) {
	pFrameALVC = (FRAMEAL_VC_T *) VC_HNDL_DATA(pBufDesc)->FrameALVCMod;
	ADD_TO_CTR(pFrameALVC->Stats.TXATMCellsMgmCell, 1);
	pFrameALVC = NULL;
      }

      OOB_DATA(pBufDesc).b.OAM_CELL = 1;
      VC_HNDL_DATA(pBufDesc) = NULL;

      ADD_TO_CTR(pFrameALLink->Stats.TXATMCellsMgmCell, 1);
    } else {
      CDSL_VC_T		*pVCHdl;
      ATM_HEADER_T	*pCellHeader;

#if FRAMEALATM_TRAFFIC_SHAPING
      if (pFrameALLink->ATMLinkBlock.TrafficShapingActive == TRAFFIC_SHAPING_OFF)
#endif 
	{
	  DWORD 		VCIndex;

	  for (VCIndex = 0; VCIndex < MAX_VC_PER_LINK; VCIndex++) {
	    pVCHdl = &LinkHdl->VC_Info[pFrameALLink->ATMLinkBlock.CurrentVC];

	    if (pFrameALLink->ATMLinkBlock.CurrentVC < (MAX_VC_PER_LINK - 1))
	      pFrameALLink->ATMLinkBlock.CurrentVC++;
	    else
	      pFrameALLink->ATMLinkBlock.CurrentVC = 0;

	    pFrameALVC = (FRAMEAL_VC_T *) pVCHdl->FrameALVCMod;

	    if (pFrameALVC && pFrameALVC->ATMVCBlock.TxGetFullBuf) {
	      pFrameALVC->ATMVCBlock.InVCALCNXT[FRAMEALVC_CNXT_GETTX] = TRUE;
	      // TxGetFullBuf returns a pointer but it is the same as pBufDesc
	      (*pFrameALVC->ATMVCBlock.TxGetFullBuf)(pVCHdl, pBufDesc);
	      pFrameALVC->ATMVCBlock.InVCALCNXT[FRAMEALVC_CNXT_GETTX] = FALSE;
	    }

	    if (pBufDesc->len > 0)
	      break;
	  }

	  if (pBufDesc->len) {
	    VC_HNDL_DATA(pBufDesc) = pVCHdl;

	    goto CCCC;
	  }
	} else {
	  FrameATMShGetCell(LinkHdl, pBufDesc);

	  if (!pBufDesc->len) {
#if 0
	    CELLHEAD_DATA(pBufDesc) = &pFrameALLink->ATMLinkBlock.IdleHeader;
	    CELLHEAD_LEN(pBufDesc) = CELL_HEADER_SIZE;

	    pBufDesc->data = &pFrameALLink->ATMLinkBlock.IdlePayload[0];
	    pBufDesc->len = CELL_PAYLOAD_SIZE;

	    OOB_DATA(pBufDesc).v = 0;
	    OOB_DATA(pBufDesc).b.IDLE = 1;

	    VC_HNDL_DATA(pBufDesc) = NULL;
#endif
	  } else {
	    pVCHdl = VC_HNDL_DATA(pBufDesc);

	    pFrameALVC = (FRAMEAL_VC_T *) pVCHdl->FrameALVCMod;

	    //VC_HNDL_DATA(pBufDesc) = pVCHdl;

	  CCCC:
	    FrameALBldHdr(pFrameALVC,pBufDesc,OOB_DATA(pBufDesc));

	    ADD_TO_CTR(pFrameALVC->Stats.TXBytes,
		       CELLHEAD_LEN(pBufDesc) + pBufDesc->len);
	    ADD_TO_CTR(pFrameALVC->Stats.COTXBytes,
		       CELLHEAD_LEN(pBufDesc) + pBufDesc->len);

	    // update VC stats
	    pCellHeader = CELLHEAD_DATA(pBufDesc);

	    ADD_TO_CTR(pFrameALVC->Stats.TXATMCellsCLP[pCellHeader->CLP], 1);
	    ADD_TO_CTR(pFrameALVC->Stats.COTXATMCells, 1);
	  }
	}
    }
  }
}

/******************************************************************************
FUNCTION NAME:
	FrameALATMGetTXCell

ABSTRACT:
	Traffic Shaper calls this function to a Cell from a VC.

RETURN:
	None

DETAILS:
	This function gets a buffer containing one ATM cell from SAR.

******************************************************************************/
GLOBAL void FrameALATMGetTXCell(
    IN	CDSL_VC_T	*VCHdl,		// VC Handle to data stream
    IN	SK_BUFF_T	*BufferDescPtr	// Pointer to a buffer descriptor
) {
  FRAMEAL_VC_T		*pFrameALVC = (FRAMEAL_VC_T *) VCHdl->FrameALVCMod;

  if (pFrameALVC && pFrameALVC->ATMVCBlock.TxGetFullBuf) {
    pFrameALVC->ATMVCBlock.InVCALCNXT[FRAMEALVC_CNXT_GETTX] = TRUE;
    // 	 TxGetFullBuf returns a pointer but it is the same as pBufDesc
    (*pFrameALVC->ATMVCBlock.TxGetFullBuf)(VCHdl, BufferDescPtr);

    if (BufferDescPtr->len > 0)
      VC_HNDL_DATA(BufferDescPtr) = VCHdl;

    pFrameALVC->ATMVCBlock.InVCALCNXT[FRAMEALVC_CNXT_GETTX] = FALSE;
  }
}

/******************************************************************************
FUNCTION NAME:
	FrameALATMStartTx

ABSTRACT:
	Calls ChipAL to start sending Cells on a Link.

RETURN:
	NDIS status

DETAILS:

******************************************************************************/
GLOBAL NDIS_STATUS FrameALATMStartTx(
    IN	CDSL_LINK_T	*LinkHdl		// Link Handle to data stream
) {
  if (LinkHdl) {
    FRAMEAL_LINK_T  	*pFrameALLink = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;

    if (pFrameALLink) {
      // if link state is idle then start the ChipAL
      if (pFrameALLink->ATMLinkBlock.LinkState == ATM_LINK_STATE_IDLE) {
	// make sure chipallink init has occurred first by checking for context
	if (LinkHdl->pThisAdapter->pChipALRemoteInf->LinkInterface.iface.Context)
	  ChipALStartTx(LinkHdl);
      }
      return STATUS_SUCCESS;
    }
  }

  return STATUS_FAILURE;
}

/******************************************************************************
FUNCTION NAME:
	FrameALATMTimerHandler

ABSTRACT:
	Calls when Hobbs timer expires.

RETURN:
	None

DETAILS:

******************************************************************************/
GLOBAL void FrameALATMTimerHandler(
    IN	CDSL_LINK_T	*LinkHdl		// Link Handle to data stream
) {
#if FRAMEALATM_TRAFFIC_SHAPING
  if (LinkHdl && LinkHdl->FrameMod) {
    FRAMEAL_LINK_T	*pFrameALLink = (FRAMEAL_LINK_T *) LinkHdl->FrameMod;
    BOOLEAN		NeedStartTx = FALSE;
    FRAMEAL_SHAPER_STATUS	OldTrafficShapingFlag;
    FRAMEAL_SHAPER_STATUS	TempTrafficShapingFlag;

    // turn off traffic shaping while updating the ATM layer
    OldTrafficShapingFlag = pFrameALLink->ATMLinkBlock.TrafficShapingActive;
    pFrameALLink->ATMLinkBlock.TrafficShapingActive = TRAFFIC_SHAPING_OFF;

    // Add this VC to the traffic shaper
    TempTrafficShapingFlag = FrameATMShTimerHandler(LinkHdl, &NeedStartTx);
    pFrameALLink->ATMLinkBlock.TrafficShapingActive = 
      (TempTrafficShapingFlag == TRAFFIC_SHAPING_NOCHANGE) ?
      OldTrafficShapingFlag : TempTrafficShapingFlag;
  }
#endif
  FrameALLinkStartTX(LinkHdl);
}

/******************************************************************************
FUNCTION NAME:
	FrameALATMLinkUp

ABSTRACT:
	This function is called when the line enters showtime

RETURNS:
	None

DETAILS:

******************************************************************************/
GLOBAL void FrameALATMLinkUp(
    IN	CDSL_LINK_T		*LinkHdl	// Link Handle to data stream
) {
#if 0/*CLN*/
  printk("CnxADSL %s:\n", __FUNCTION__);
#endif
  FrameALATMOAMLinkUp(LinkHdl);
}

/******************************************************************************
FUNCTION NAME:
	FrameALLinkDown

ABSTRACT:
	This function is called when the line leaves showtime

RETURNS:
	None

DETAILS:

******************************************************************************/

GLOBAL void FrameALATMLinkDown(
    IN	CDSL_LINK_T		*LinkHdl	// Link Handle to data stream
) {
#if 0/*CLN*/
  printk("CnxADSL %s:\n", __FUNCTION__);
#endif
}
